classdef TruncatedInfo < Fem.Lagrange.DefaultInfo
    %classdef TruncatedInfo < DefaultInfo
    % truncated scalar Lagrange FE functions on triangular grid
    
    methods
        
        function truncated_scalar_df_info = TruncatedInfo(model, scalar_df_info, bc_info)
            % Check not implemented cases
            if ~strcmp(model.gridtype, 'triagrid')
                % Hint: '' in a string prints an apostroph (') in the string which
                % means that ''' prints one apostroph in the string and ends the string
                error('TruncatedInfo is yet only implemented for model.gridtype=''triagrid''');
            end
            if bc_info.df_info.dimrange ~= 2
                error('TruncatedInfo is yet only implemented for dimrange = 2');
            end
            
            full_grid = scalar_df_info.grid;
            
            % Extract fully constrained Dirichlet nodes from grid if neccessary
            if model.has_dirichlet_values
                % Extract p and t from p-e-t formulation
                p = [full_grid.X'; full_grid.Y'];
                t = full_grid.VI';
                nvert = size(p,2);

                % Fetch Dirichlet nodes indices
                idx_diri = bc_info.dirichlet_gids;

                % Determine fully constrained nodes (x- and y-displacement)
                idx_fully_constr = mod(idx_diri-1, nvert)+1;

                % Check, if only fully constrainted nodes exist
                if mod(numel(idx_fully_constr),2) ~= 0 || ...
                        norm(idx_fully_constr(1:end/2) - idx_fully_constr(end/2+1:end))
                    % The reason for this error is that if the comfortable way to
                    % remove the whole node for all Dirichlet values does not work if
                    % there are nodes that are constrained only in X or only in Y
                    % direction
                    error('TruncatedInfo is only implemented for a model where all Dirichlet nodes are constrained in x AND y direction!');
                end

                % Extract fully constrained nodes
                idx_fully_constr = idx_fully_constr(1:end/2);

                % Remove fully constrained nodes from p
                p(:, idx_fully_constr) = [];

                % Remove elements with fully constrained nodes from t
                idx_tria_to_remove = any(ismember(t,idx_fully_constr),1);
                t(:, idx_tria_to_remove) = [];

                % Shift all vertex indices by deleted vertices
                idx_fully_constr = sort(idx_fully_constr(:), 'descend');
                for ii=1:numel(idx_fully_constr)
                    t = t - (t>=idx_fully_constr(ii));
                end

                % Construct grid from truncated p-e-t formulation
                grid = triagrid(p,t,model);
                grid = set_boundary_types(grid, model);
            else
                grid = full_grid;
            end
            
            truncated_scalar_df_info = truncated_scalar_df_info@Fem.Lagrange.DefaultInfo(model.pdeg, grid);
            
            % Reset truncated operators stored in DefaultInfo if neccessary
            if model.has_dirichlet_values
                % Get and truncated l2 inner product matrix
                truncated_scalar_df_info.l2_inner_product_matrix = scalar_df_info.l2_inner_product_matrix;
                truncated_scalar_df_info.l2_inner_product_matrix(idx_fully_constr, :) = [];
                truncated_scalar_df_info.l2_inner_product_matrix(:, idx_fully_constr) = [];
                % Get and truncated h10 inner product matrix
                truncated_scalar_df_info.h10_inner_product_matrix = scalar_df_info.h10_inner_product_matrix;
                truncated_scalar_df_info.h10_inner_product_matrix(idx_fully_constr, :) = [];
                truncated_scalar_df_info.h10_inner_product_matrix(:, idx_fully_constr) = [];
            end
        end
    end
end


