classdef OperatorsTruncated < Fem.OperatorsDefault
    %classdef OperatorsTruncated < Fem.OperatorsDefault
    % Truncated Fem operatos can be obtained from stored components.
    %
    % Special attention needs to be spend since there are two models and
    % model_datas: the truncated and the full system
    % 
    % Dirichlet values are truncated.
    
    properties
    end
    
    methods
        function this = OperatorsTruncated()
            %function this = OperatorsTruncated()
            % Only initializing new object.
        end
        
        function varargout = evaluate(this, model, model_data)
            %function varargout = evaluate(this, model, model_data)
            % Operator evaluation. Returns a truncated Fem system.
            %
            % If the components are stored, the system is computed via
            % linear combinations of the components, else the assembly
            % routine of model is called.
            %
            % Return values:
            %  varargout: The first output represents the system matrix,
            %  the second output (only if requested) the right-hand side
            %  vector. The field model.decomp_mode has no influence.
            
            if ~model.has_dirichlet_values
                varargout = evaluate@Fem.OperatorsDefault(this, model, model_data);
            end
            
            if this.componentsAssembled
                model.decomp_mode = 2;
                [A_coeff, f_coeff] = model.operators(model, model_data);
                
                if model.has_nonlinearity
                    % Reason: Fem.OperatorsDefault. Have to look at Qr_dir
                    % stuff
                    error('Models with nonlinearity are not implemented yet with componentsAssembled!');
                end
                
                varargout{1} = lincomb_sequence(this.A_comp, A_coeff);
                
                if nargout > 1
                    % Linear combinateion of RHS
                    %   Dirichlet values are already included due to
                    %   implementation of Fem.assembly.operator()
                    %   Compare with OperatorsDefault. There the r_dir
                    %   coefficients in f_coeff are set to zero because the
                    %   system does not get truncated
                    varargout{2} = lincomb_sequence(this.f_comp, f_coeff);
                end
            else
                model.decomp_mode = 0;
                [varargout{1:nargout}] = model.operators(model, model_data);
                
                % Get indices of Dirichlet dofs
                dirichlet_gids = model_data.bc_info.dirichlet_gids;
                
                if nargout > 1
                    if iscell(model_data.bc_info.dirichlet_dof_vector_components)
                        model.decomp_mode = 2;
                        r_dir_coeff = model.dirichlet_values([], [], [], [], model);
                        model.decomp_mode = 0;

                        r_dir = lincomb_sequence(model_data.bc_info.dirichlet_dof_vector_components, ...
                            r_dir_coeff);
                    else
                        r_dir = model_data.bc_info.dirichlet_dof_vector_components;
                    end
                    
                    % Undo changes from Fem.Assembly.operator
                    varargout{2} = varargout{2} - r_dir;

                    % Add Dirichlet values to RHS
                    varargout{2} = varargout{2} - varargout{1}*r_dir;
                    
                    % Truncation of Dirichlet dofs in RHS
                    varargout{2}(dirichlet_gids) = [];
                end
                
                % Truncation of Dirichlet dofs in system matrix
                varargout{1}(dirichlet_gids, :) = [];
                varargout{1}(:, dirichlet_gids) = [];
            end                    
        end
        
        function this = assemble_components(this, model, model_data)
            this = assemble_components@Fem.OperatorsDefault(this, model, model_data);
            
            if model.has_dirichlet_values
                % Get indices of Dirichlet dofs
                dirichlet_gids = model_data.bc_info.dirichlet_gids;
                
                % Truncation of Dirichlet dofs in A
                Q_A = length(this.A_comp);
                for q=1:Q_A
                    this.A_comp{q}(dirichlet_gids, :) = [];
                    this.A_comp{q}(:, dirichlet_gids) = [];
                end
                
                % Truncation of Dirichlet dofs in RHS
                Q_f = length(this.f_comp);
                for q=1:Q_f
                    this.f_comp{q}(dirichlet_gids) = [];
                end
            end                    
        end
    end
    
end

