function [t,y] = implicit_midpoint(ode, tspan, dt, y0, jac)
% function [t,y] = implicit_midpoint(ode, tspan, y0, jac)
% solves first order ode y' = f(t,y) in the time interval tspan with initial condition
% y0 with the (symplectic) implicit midpoint rule
%   Fully implicit: if jacobian is provided
%   Ecplicit, iterative sheme: otherwise
%
% A. Bhatt, April 2017
% Patrick Buchfink, June 2017

t = tspan(1):dt:tspan(2);
N = length(t);
At = @(x,y) (x+y)/2;
tol = 10^-15;
n_dof = size(y0,1);
y = zeros(n_dof, N);
y(:,1) = y0;

% right hand side for implicit midpoint
rhs = @(t,y) y(:,1) + dt*ode(At(t(1),t(2)),At(y(:,1),y(:,2)));

% fixed point iteration
if nargin < 5
    for n = 1:N-1
        yk = rhs([t(n),t(n+1)],[y(:,n),y(:,n)]);

        while 1
            y(:,n+1) = rhs([t(n),t(n+1)],[y(:,n),yk]);

            if norm(yk-y(:,n+1),inf) < tol
                break
            end
            yk = y(:,n+1);
        end
        
        % Print progress in multiples of 10%
        if floor(((n-1)/N)*10) < floor((n/N)*10)
            fprintf('implicit midpoint: %3d%%\n', floor((n/N)*10)*10);
        end
    end
    % Print progress of 100%
    fprintf('implicit midpoint: %3d%%\n', 100);
    
% Fully implicit (for linear systems)
else
    % Check, if Jacobian is non-constant (i.e. system is non-linear)
    if isa(jac, 'function_handle')
        error('Implicit midpoint scheme only implemented for linear systems yet!');
    end
    
    I = speye(size(jac));
    
    for n = 1:N-1
        % the formulation
        %   y(:,n) + (I-dt/2*jac)\(dt * ode(At(t(n),t(n+1)), y(:,n)));
        % is much more stable than
        %   y(:,n) + dt * (I-dt/2*jac)\(ode(At(t(n),t(n+1)), y(:,n)));
        y(:,n+1) = y(:,n) + (I-dt/2*jac)\(dt * ode(At(t(n),t(n+1)), y(:,n)));
        
         % Print progress in multiples of 10%
        if floor(((n-1)/N)*10) < floor((n/N)*10)
            fprintf('fully implicit midpoint: %3d%%\n', floor((n/N)*10)*10)
        end
    end
    % Print progress of 100%
    fprintf('fully implicit midpoint: %3d%%\n', 100);
end

y = y';