function [t,y] = rescaled_implicit_midpoint(ode, tspan, dt, y0, M, K)
% function [t,y] = rescaled_implicit_midpoint(tspan, dt, y0, M, K, f)
% An implementation explicitly for linear elasticity in the (q,p)
% formulation. In difference to the implicit_midpoint, it avoids the
% explict usage of inv{M}.
% 
% Patrick Buchfink, November 2017

t = tspan(1):dt:tspan(2);
N = length(t);
At = @(x,y) (x+y)/2;
n_dof = size(y0,1);
y = zeros(n_dof, N);
y(:,1) = y0;

% Check, if stiffness matrix is non-constant (i.e. non-linear elasticity)
if isa(K, 'function_handle')
    error('Implicit midpoint scheme only implemented for linear elasticity yet!');
end

n  = n_dof/2;
In = speye(n,n);

% rescale matrix in front of LHS
A = [In    , -dt/2 * In; ...
     dt/2*K,          M];
S = blkdiag(In, M);

for n = 1:N-1
    % add increment
    y(:,n+1) = y(:,n) + S*( A\(dt * ode(At(t(n),t(n+1)), y(:,n))) );

     % Print progress in multiples of 10%
    if floor(((n-1)/N)*10) < floor((n/N)*10)
        fprintf('rescaled implicit midpoint: %3d%%\n', floor((n/N)*10)*10)
    end
end
% Print progress of 100%
fprintf('rescaled implicit midpoint: %3d%%\n', 100);

y = y';