function [S, D, Q, p, q, meta_data] = svd_like_decomposition(B, init_decomp, verbosity, debug)
%function svd_like_decomposition(B, init_decomp, verbosity, debug)
% This function computes the full SVD-like decomposition
% This algorithm follows the paper
%   Xu, A numiercal method for computing an svd-like decomposition", SIAM J. MATRIXANAL.APPL Vol. 26, No. 4, pp. 1058–1082
% which is also available open-source here
%   https://pdfs.semanticscholar.org/18d3/87515c5db69a1492b41c859fcf0d659df924.pdf
%
% Input:
%   B - the matrix to be decomposed
%   init_decomp - a flag which initial decomposition to use
%   verbosity - a flag for the verbosity level of the function
%   debug - a flag whether to compute debug quantities and save them in meta_data
% 
% Output:
%   S,D,Q - matrices of an SVD-like decomposition Q'*B*S = D
%       S: square, symplectic matrix
%       D: sparse matrix
%       Q: square, orthogonal matrix
%   p - half the rank of B*J*B', i.e. rank(B*J*B') = 2*p
%   q - difference of rank(B*J*B') and rank(B), i.e. rank(B) = 2*p+q
%   meta_data - used for meta data computed if debug=1
%
% TODO: If q > 0, J2n-orthogonality of S is very bad

t_start = tic();

global verbose it_mod_verbosity f_debug abs_tol;
if nargin < 4
    debug = 1;
    if nargin < 3
        verbosity = 1;
        if nargin < 2
            init_decomp = 'svd';
        end
    end
end
verbose = verbosity;
f_debug = debug;
it_mod_verbosity = 50;

abs_tol = eps * norm(B);

n = size(B, 1);
m = size(B, 2) / 2;

%% Treatment for the general case

% rank of B (is not necessarily equal to p!)
r = rank(B);

% initial decomposition such that B0 has full row rank r
switch(init_decomp)
    case 'svd'
        [Q, SigS, VS] = svd(B);
        B0 = SigS(1:r,1:r)*VS(:,1:r)';
    case 'qr'
        [Q, R, P] = qr(B);
        B0 = R(1:r, :) * P';
    otherwise
        error('Unkown initial decomposition: %s', init_decomp);
end

% sanity check
if rank(B0) - size(B0, 1) > 0
    error('B0 is supposed to have full row rank!');
end

% recompute truncated B if debugging is enabled
if debug > 0
     B = Q * [B0; zeros(n-r, 2*m)];
end

%% Step 1
% reduction of B0 with Q_2^T * B_0 * U_2
[QT_B_U, Q2T, U2, p, q] = implicit_reduce(B0);
B11 = QT_B_U(1:p, 1:p);
B34 = QT_B_U(p+q+(1:p), m+(1:p));

%% Step 2
% implicit QR-like SVD iteration (Step 2)
[Z1T, Z2T, W] = implicite_QR_like_SVD_iteration(B11, B34);

%% Step 3
% is skipped since we are not interested in the eigenvalues of J*B^T*B or B*J*B^T

%% Step 4
% Compute Q^T * B * U
QT_B_U(1:p,:) = Z1T * QT_B_U(1:p,:);
QT_B_U(p+q+(1:p),:) = Z2T * QT_B_U(p+q+(1:p),:);
QT_B_U(:,1:p) = QT_B_U(:,1:p) * W;
QT_B_U(:,m+(1:p)) = QT_B_U(:,m+(1:p)) * W;
% update Q2 and U2 accordingly
Q2T(1:p,:) = Z1T * Q2T(1:p,:);
Q2T(p+q+(1:p),:) = Z2T * Q2T(p+q+(1:p),:);
U2(:, 1:p) = U2(:, 1:p) * W;
U2(:, m+(1:p)) = U2(:, m+(1:p)) * W;

% switch all diagonal entries of R11 and R34 to positive signs
idx_neg = find(diag(QT_B_U(1:p, 1:p)) < 0);
QT_B_U(idx_neg, :) = -QT_B_U(idx_neg, :);
Q2T(idx_neg, :) = -Q2T(idx_neg, :);
idx_neg = p+q+ find(diag(QT_B_U(p+q+(1:p), m+(1:p))) < 0);
QT_B_U(idx_neg, :) = -QT_B_U(idx_neg, :);
Q2T(idx_neg, :) = -Q2T(idx_neg, :);

% Extract R terms
% TODO: directly insert R terms by index
R11 = QT_B_U(1:p, 1:p);
R12 = QT_B_U(1:p, p+(1:q));
R22 = QT_B_U(p+(1:q), p+(1:q));
R13 = QT_B_U(1:p, p+q+(1:m-p-q));
R14 = QT_B_U(1:p, m+(1:p));
R15 = QT_B_U(1:p, m+p+(1:q));
R16 = QT_B_U(1:p, m+p+q+(1:m-p-q));
R34 = QT_B_U(p+q+(1:p), m+(1:p));

Delta = R11*R34';

if norm(Delta - diag(diag(Delta))) > abs_tol*norm(Delta) || any(diag(Delta) < 0)
    error('Delta is supposed to be a positive diagonal matrix!');
end

Sigma = diag(sqrt(diag(Delta)));

% sort by magnitude of Sigma
[~, idx_perm] = sort(diag(Sigma), 'desc');
Sigma = Sigma(idx_perm, idx_perm);

invSigma = diag(1./diag(Sigma));

X = R34(idx_perm, :)' * invSigma;

S = [
     [X; zeros(2*m-p, p)], ... first block column
     [... begin second block column
      -X*(R12(idx_perm, :)' * invSigma)'; ...
      eye(q); ...
      zeros(2*m-p-q, q)
     ], ... end second block column
     [... begin third block column
      -X*(R13(idx_perm, :)' * invSigma)'; ...
      zeros(q, m-p-q); ...
      eye(m-p-q); ...
      zeros(m, m-p-q)
     ], ... end third block column
     [... begin fourth block column
      -R14(idx_perm, :)' * invSigma; ...
      -R15(idx_perm, :)' * invSigma; ...
      -R16(idx_perm, :)' * invSigma; ...
       R11(idx_perm, :)' * invSigma; ...
       R12(idx_perm, :)' * invSigma; ...
       R13(idx_perm, :)' * invSigma ... 
     ],... end fourth block column
     [... begin fifth block column
      -X*(R15(idx_perm, :)' * invSigma)'; ...
      zeros(m, q); ...
      eye(q); ...
      zeros(m-p-q, q)
     ], ... end fifth block column
     [... begin sixth block column
      -X*(R16(idx_perm, :)' * invSigma)'; ...
      zeros(m+q, m-p-q); ...
      eye(m-p-q) ...
     ] ... end sixth block column
    ];

S = U2 * S;
S(:,p+(1:q)) = S(:,p+(1:q)) / R22;
S(:,m+p+(1:q)) = S(:,m+p+(1:q)) * R22';

Q2T(1:p, :) = Q2T(idx_perm, :);
Q2T(p+q+(1:p), :) = Q2T(p+q+idx_perm, :);

Q(:, 1:size(B0, 1)) = Q(:, 1:size(B0, 1)) * Q2T';

D = sparse(n, 2*m);
D(1:p,1:p) = Sigma;
D(p+(1:q), p+(1:q)) = spdiags(ones(q,1),0,q,q);
D(p+q+(1:p), m+(1:p)) = Sigma;

t_runtime = toc(t_start);

if verbose > 0
    fprintf('SVD-like decomposition finshed after: %4.2e s\n', t_runtime);
end

if f_debug > 0
        J2 = @(n) [sparse(n, n), speye(n, n); ...
                  -speye(n, n), sparse(n, n)];
        J2m = J2(m);
        meta_data.err = norm(Q*D*J2m'*S'*J2m - B);
        meta_data.err_inv = norm(Q'*B*S - D);
        meta_data.ortho = norm(Q'*Q - eye(n));
        meta_data.sympl = norm(S'*J2m*S - J2m);
end

end

function [B0, Q2T, U2, p, q] = implicit_reduce(B0)

global verbose it_mod_verbosity abs_tol;

if verbose > 0
    fprintf('start implicit reduction\n');
end

n = size(B0, 1);
p = floor(n/2);
m = size(B0, 2) / 2;

J2 = @(n) [sparse(n, n), speye(n, n); ...
           -speye(n, n), sparse(n, n)];
J2m = J2(m);

j = 0;
q = 0;
r = n - p;

U2 = eye(2*m);
Q2T = eye(n);

it = 0;

%% step 1: condensed form (4.4)
while(j < p || r > p)
    
    % orthogonal symplectic transformations
    % annihilate B0(p+q+j+1, j+2:m-q)
    Uj1 = sympl_house(B0(p+q+j+1, j+1:m-q), j+1, m);
    B0 = Uj1(B0);
    U2 = Uj1(U2);
    % annihilate B0(p+q+j+1, j+1)
    Vj = sympl_givens_1(B0(p+q+j+1, :), j+1, m);
    B0 = Vj(B0);
    U2 = Vj(U2);
    % annihilate B0(p+q+j+1, m+j+2:2*m-q)
    Uj2 = sympl_house(B0(p+q+j+1, m+j+1:2*m-q), j+1, m);
    B0 = Uj2(B0);
    U2 = Uj2(U2);
    
    % Householder transformation
    % annihilate B0([j+2:p, p+q+j+2:p+q+r], j+1)
    Q = house_special(B0([j+1:p, p+q+j+2:p+q+r], j+1), [j+1:p, p+q+j+2:p+q+r],  j+1);
    B0 = Q(B0);
    Q2T = Q(Q2T);

    % branching two cases a. and b.
    if abs(B0(j+1, j+1)) > abs_tol
        % Householder transform
        % annihilate [B0*J*B0']([j+2:p, p+q+j+3:p+q+r], j+1)
        % care: p+q+j+2 is the non-zero entry!, not j+2
        Q = house_special(B0([j+2:p, p+q+j+2:p+q+r], :) *J2m* B0(j+1, :)', [j+2:p, p+q+j+2:p+q+r],  p+q+j+2);
        B0 = Q(B0);
        Q2T = Q(Q2T);

        % increment j
        j = j + 1;

    else
        % deflate zero eigenvalue of B0*J*B0^T
        % (sub-cases are the same until the row permutation step)
        B0(j+1, j+1) = 0;
        
        % Symplectic Givens transformations
        % annihilate diagonal elements in B0(1:j, 1:j)
        for i=j:-1:1
            Gji = sympl_givens_2(B0(i, i+(0:1)), i, i+1, m);
            B0 = Gji(B0);
            U2 = Gji(U2);
        end
        
        % Givens transformations
        % annihilate diagonal elements in B0(p+q+(1:j), m+1+(1:j))
        for i=j:-1:i
            Gji = givens_special(B0(p+q+i+(1:-1:0), m+i+1), p+q+i+1, p+q+i);
            B0 = Gji(B0);
            Q2T = Gji(Q2T);
        end
        
        % symplectic permutation (columns)
        Pj1 = sympl_move(1, m-q, m);
        B0 = Pj1(B0);
        U2 = Pj1(U2);
        Pj2 = sympl_switch(m-q, m);
        B0 = Pj2(B0);
        U2 = Pj2(U2);

        % branch two cases (b1) and (b2)
        if r == p+1
            
            % symplectic permotation (rows)
            Pj3 = move_special(p+q+1, p+1, n);
            B0 = Pj3(B0);
            Q2T = Pj3(Q2T);
            
            r = r-1;
            q = q+1;
        elseif r == p
            % symplectic permutation (rows)
            % move row p -> n
            Pj3 = move_special(p, n, n);
            B0 = Pj3(B0);
            Q2T = Pj3(Q2T);
            % move row p+q -> p
            Pj3 = move_special(p+q, p, n);
            B0 = Pj3(B0);
            Q2T = Pj3(Q2T);
            
            p = p-1;
            q = q+1;
        else
            error('This case is not defined!');
        end
    end
    
    it = it+1;
    if verbose > 0 && mod(it, it_mod_verbosity) == 0
        fprintf('impl. red., it: %d, p: %d, q: %d, r: %d\n', it, p, q, r);
    end
        
end

%% step 2: orthogonal symplectic transformations (4.4) -> (4.2)

% symplectic Givens transformations
% to annihilate B0(p+q+(1:q), m-q+(1:q))
for j=1:q
    for i=1:p
        Gji = sympl_givens_3(B0(p+q+i, [m+i, m-q+j]), i, m-q+j, m);
        B0 = Gji(B0);
        U2 = Gji(U2);
    end
end

% symplectic permutation (columns)
% move the columns p+(1:m-p-q) to the position m+(-(m-p-q):-1)
idx_perm = [1:p, m-q+(1:q), p+1:m-q];
B0 = B0(:, [idx_perm, m+idx_perm]);
U2 = U2(:, [idx_perm, m+idx_perm]);

if verbose > 0
    fprintf('finished implicit reduction, it: %d, p: %d, q: %d, r: %d\n', it, p, q, r);
end

end

function [Z1T, Z2T, W] = implicite_QR_like_SVD_iteration(B11, B34)

global verbose it_mod_verbosity;

if verbose > 0
    fprintf('start implicit QR-like SVD iteration\n');
end

p = size(B11, 1);

Z1T = eye(p);
Z2T = eye(p);
W  = eye(p);

%% lower Hessenberg form for B34
% such that B11_B34T is lower bidiagonal

for j=1:p-1
    Gj = givens_special(B11(j, :) * B34([j, j+1], :)', j, j+1);
    B34 = Gj(B34);
    Z2T = Gj(Z2T);
end

%% apply implicit shift QR algorithm for each block

k = 0;
it = 0;
while k < p
    
    % set near zero super-diagonal entries to zero (criterion from (3.1))
    idx_zero_supdiag = find( ...
        abs(diag(B34, 1)) <= eps * max(abs(diag(B34(1:p-1,1:p-1))) + abs(diag(B34, -1)) + abs(diag(B34(2:p, 2:p))), 1) );
    B34((p+1) * idx_zero_supdiag) = 0;
    
    % apply unreduced_step to unreduced blocks
    idx_zero_supdiag = [0, idx_zero_supdiag(:)', p];
    k = numel(idx_zero_supdiag)-1;
    for i_block = 1:k
        idx_op = idx_zero_supdiag(i_block)+1:idx_zero_supdiag(i_block+1);
        
        if numel(idx_op) > 1
            [Z1T_up, Z2T_up, W_up] = ...
                unreduced_step(B11(idx_op, idx_op), B34(idx_op, idx_op));

            % update B matrices
            B11(idx_op, :) = Z1T_up * B11(idx_op, :);
            B11(:, idx_op) = B11(:, idx_op) * W_up;
            B34(idx_op, :) = Z2T_up * B34(idx_op, :);
            B34(:, idx_op) = B34(:, idx_op) * W_up;

            Z1T(idx_op, :) = Z1T_up * Z1T(idx_op, :);
            Z2T(idx_op, :) = Z2T_up * Z2T(idx_op, :);
            W(:, idx_op)  = W(:, idx_op) * W_up;
        end
    end
    
    it = it+1;
    if verbose > 0 && mod(it, it_mod_verbosity) == 0
        idx_last = idx_zero_supdiag(end-1);
        fprintf('impl. SVD it., it: %d, p: %d, k: %d, last enrty: %4.2e\n',...
            it, p, k, B34(idx_last, idx_last+1));
    end
    if it > intmax / 2
        error('Too many iterations in svd-like decomposition!');
    end
end

if verbose > 0
    fprintf('finished implicit QR-like SVD iteration\n');
end

end

function [Z1T, Z2T, W] = unreduced_step(B11, B34)
% partly following Algorithm: 8.6.1 from Matrix Computations, Golub & Van Loan
% TODO: implement efficient storage of bidiagnoal matrix B suggested by Golub

n = size(B11, 1);

Z1T = eye(n,n);
Z2T = eye(n,n);
W = eye(n,n);

% compute trailing and leading 2x2 block in T =  B11 * (B34' * B34) * B11'
trailing_2x2_T = B11(n-1:n, :) * B34';
trailing_2x2_T = trailing_2x2_T * trailing_2x2_T';

leading_2x2_T = B11(1:2, :) * B34';
leading_2x2_T = leading_2x2_T * leading_2x2_T';

% compute Wilkinson shift
d = (trailing_2x2_T(1,1) - trailing_2x2_T(2,2))/2;
mu = trailing_2x2_T(2,2) - trailing_2x2_T(2,1)^2 / ...
    (d + sign(d) * sqrt(d^2 + trailing_2x2_T(2, 1)^2));

% initial modifications k=0
Gk = givens_special([leading_2x2_T(1,1) - mu, leading_2x2_T(2,1)], 1, 2);
B11 = Gk(B11);
Z1T = Gk(Z1T);

% annihilate B11(2, 1)
Wk = givens_special(B11(2, [2, 1]), 2, 1, 'right');
B11 = Wk(B11);
B34 = Wk(B34);
W = Wk(W);
    
% annihilate B11_B34T(1, 2)
Sk = givens_special(B11(1, :) * B34([1,2], :)', 1, 2);
B34 = Sk(B34);
Z2T = Sk(Z2T);

for k=2:n-1
    
    % annihilate B34(k-1, k+1)
    Wk = givens_special(B34(k-1, [k, k+1]), k, k+1, 'right');
    B11 = Wk(B11);
    B34 = Wk(B34);
    W = Wk(W);

    % annihilate B11(k+1, k)
    Gk = givens_special(B11([k, k+1], k), k, k+1);
    B11 = Gk(B11);
    Z1T = Gk(Z1T);
    
    % annihilate B11_B34T(k, k+1)
    Sk = givens_special(B11(k, :) * B34([k, k+1], :)', k, k+1);
    B34 = Sk(B34);
    Z2T = Sk(Z2T);
end

end


