function model = beam_RBM_model(input_params)
% A realization of dyn_lin_elast_model and quadratic_symplectic_model
%   with a (rectangular) beam geometry and a time-dependent Dirichlet value
%
%   As the boundary traction is time-dependent, the model is
%   non-autonomous
%
% Patrick Buchfink, 2018

%% preallocate input parameters
if nargin < 1
    input_params = struct();
end

if ~isfield(input_params, 'beam_motion')
    input_params.beam_motion = 'sinusoidal';
end

if ~isfield(input_params, 'use_nondim')
    input_params.use_nondim = false;
end

if ~isfield(input_params, 'default_mu')
    % very low stiffness for fun pictures
    input_params.default_mu = [50,50,1];
end

if ~isfield(input_params, 'default_mu_range')
    input_params.default_mu_range = {[50 50], [50 50], [1,1]};
end

%% construct a Hamiltonian beam model
if input_params.use_nondim
    % prevent the model to apply non-dimensionalization too early
    input_params.apply_non_dim = false;
    % create beam model without non-dimensionalization
    model = beam_autonomous_model(input_params);
else
    model = beam_autonomous_model(input_params);
end

%% changes in the dynamical, linear elastic model
% add time-dependence in the Dirichlet values coefficients
%   important! set second-order time-derivative of Dirichlet values
%   coefficients to match the Dirichlet values coefficients
switch(input_params.beam_motion)
    % sinusoidal movement of the Dirichlet DOFs in y direction
    case 'sinusoidal'
        model.dirichlet_values_coefficients = {
            @(params) 0, ...
            @(params) sine_RBM(params)};
        model.omega = 1;
        
        model.ddt_dirichlet_values_coefficients = {...
            @(params) 0, ...
            @(params) -params.omega^2 * sin(params.omega * params.t)};
    
    % constant motion in y direction
    case 'constant'
        model.dirichlet_values_coefficients = {
            @(params) 0, ...
            @(params) constant_RBM(params)};
        model.const_velocity = 1;
        
        model.ddt_dirichlet_values_coefficients = {...
            @(params) 0, ...
            @(params) 0};
        
    otherwise
        error('Unknown input_params.beam_motion: %s', input_params.beam_motion);
end
model.dirichlet_values_components   = @(grid, eindices, faceind, llcoord, params) ...
    {rbm_x_dir(grid, eindices, faceind, llcoord, params), ...
    rbm_y_dir(grid, eindices, faceind, llcoord, params)};

% remove Neumann boundary force
model.neumann_values_coefficients = {@(params) 0};
model.neumann_values_components   = @(grid, eindices, faceind, llcoord, params) ...
    {zeros(numel(eindices),2)};

% update coefficients
model = model.update_coefficients(model);

%% changes in the symplectic framework
% change to a time-depenendent vector in the Hamiltonian
model.h = @(t,M,K,f) [-f(t); zeros(size(f(t)))];

%% change flags of time-dependent RHS coefficients
%   due to time-dependent Dirichlet conditions
model.has_time_dependent_rhs_coefficients = true;
model.is_h_time_dependent                 = true;

%% change flags for time-dependent Dirichlet coefficients
model.has_time_dependent_Dirichlet_coefficients = true;

%% set physical parameters
model.tspan = [0, 2*pi];
model.dt    = diff(model.tspan)/150;

%% set initial condition such that the system is in equilibrium at the initial time
if strcmp(input_params.beam_motion ,'constant')
    % initial velocity of the rigid body motion
    model.rbm_initial_velocity = @(params) [0; params.const_velocity];

    model.q0 = @(params, model_data) q0(params, model_data);
    model.v0 = @(params, model_data) v0(params, model_data);
end

%% TODO: Store M * diri_component somewhere such that MOR is easier
%   and implement MOR with time-dep dirichlet

%% apply non-dimensionalization in the end
%   use this after spezifying tspan, dt, Neumann, source and Dirichlet
%   coefficients, q0, v0, char_lame_mu, char_density, char_g
if input_params.use_nondim
    model = model.apply_nondimensionalization(model);
end

end

% Dirichlet value component with rigid body displacement in x direction
function res = rbm_x_dir(~, eindices, ~, ~, ~)
    res = [ones(numel(eindices), 1), zeros(numel(eindices), 1)];
end

% Dirichlet value component with rigid body displacement in y direction
function res = rbm_y_dir(~, eindices, ~, ~, ~)
    res = [zeros(numel(eindices), 1), ones(numel(eindices), 1)];
end

% function to compute initial value such that the whole system is in
% equilibrium at the initial time t0
function q0 = q0(model, model_data)
    % Get stiffness matrix and RHS (without boundary forces)
    model.decomp_mode = 0;
    [K, f] = model_data.operators(model, model_data);
    % Calculate equilibrium for initial forces
    q0 = K\f;

    % In the case of a dimless model, q is calculated from dimless quantities
    % (K and f are already dimless). Thus, q0 is dimless itself and it needs to
    % be rescaled to a dimensioned quantity
    if isfield(model, 'char_u')
        q0 = model.char_u * q0;
    end
end
function v0 = v0(params, model_data)
    n = model_data.truncated_df_info.ndofs; % dofs in x and y dir
    rbm_initial_velocity = params.rbm_initial_velocity(params);
    v0 = [ones(n/2, 1) * rbm_initial_velocity(1);
          ones(n/2, 1) * rbm_initial_velocity(2)];
end

function res = sine_RBM(params)
    res = sin(params.omega * params.t);
    if isempty(res)
        res = 0;
    end
end

function res = constant_RBM(params)
    res = params.const_velocity * params.t;
    if isempty(res)
        res = 0;
    end
end
    