function model = beam_autonomous_model(input_params)
% A realization of dyn_lin_elast_model and quadratic_symplectic_model
%   with a (rectangular) beam geometry and a constant boundary traction on
%   the far-right vertical boundary
%
%   As the boundary traction is constant in time, the model is autonomous
%
% Patrick Buchfink, 2018

%% preallocate input parameters
if nargin < 1
    input_params = struct();
end
if ~isfield(input_params, 'use_nondim')
    input_params.use_nondim = false;
end
if ~isfield(input_params, 'default_mu')
    input_params.default_mu = [122e9, 81e9, 7856];
end
if ~isfield(input_params, 'default_mu_range')
    input_params.default_mu_range = {[35e9 125e9], [35e9 83e9], [7856,7856]};
end

%% construct general dynamic linear elasticity model
if input_params.use_nondim
    model = dimless_dyn_lin_elast_model();
    % set characterisctic constants
    model.char_lame_mu = input_params.default_mu(2);
    model.char_density = input_params.default_mu(3);
    model.char_g       = 9.81;
else
    model = dyn_lin_elast_model();
end

%% add symplecticity framework on top
model = quadratic_symplectic_model(model);

%% fill in mandatory data in dynamic linear elasticity model
% range of the parameters mu
%   it is chosen to be two different kinds of steel
model.mu_ranges = input_params.default_mu_range;
% set default value for model (mild steel)
model = set_mu(model, input_params.default_mu);

% Neumann values coefficients and components
model.neumann_values_coefficients = {@(params) -80*9.81};
model.neumann_values_components   = @(grid, eindices, faceind, llcoord, params) ...
    {neumann_values_components(grid, eindices, faceind, llcoord, params)};
% Dirichlet values coefficients and components
model.dirichlet_values_components   = @(grid, eindices, faceind, llcoord, params) ...
    {zeros(numel(eindices), 2)};
model.dirichlet_values_coefficients = {@(params) 1};
% source term coefficients and components
model.source_components   = @(grid, eindices, loc, params) ...
    {zeros(numel(eindices), 2)};
model.source_coefficients = {@(params) 1};

% set time dependence of the rhs coefficients to false
model.has_time_dependent_rhs_coefficients = false;
model.is_h_time_dependent                 = false;
% set time dependence of the Dirichlet coefficients to false
model.has_time_dependent_Dirichlet_coefficients = false;

% set simulation time to one oscillation (with the mild steel parameters)
model.tspan = [0, 7.2e-2];
model.dt    = diff(model.tspan)/150;

% set integrator to implicit midpoint
model.integrator = 'rescaled_implicit_midpoint';

% set initial values
model.q0 = @(model, model_data) q0(model, model_data);
model.v0 = @(model, model_data) v0(model, model_data);

% apply nondimensionalization if desired
%   use this after spezifying tspan, dt, Neumann, source and Dirichlet
%   coefficients, q0, v0, char_lame_mu, char_density, char_g
if input_params.use_nondim && ...
    (~isfield(input_params, 'apply_non_dim') || ...
     (isfield(input_params, 'apply_non_dim') && input_params.apply_non_dim)...
    )
    
    model = model.apply_nondimensionalization(model);
end
% update coefficients
model = model.update_coefficients(model);

% compute RB_mu_list from a 3 samples for each lame constant and one for
% the density
model = model.create_mu_list(model, [3, 3, 1]);

% set default basis generation technique to PSD_complex_SVD with a size 20
model.MOR_technique = 'PSD_complex_SVD';
model.RB_size       = 20;

%% add geometry description for the model data
% set gird generation technique to standard rbMatlab workflow with the
% function construct_grid
model.grid_generation = 'construct_grid';
% define geometry
model.xrange = [0 8];
model.yrange = [0 1];
model.gridtype = 'triagrid';
model.xnumintervals = 64;
model.ynumintervals = 8;

% define boundaries for Dirichlet and Neumann BC
%	see Fem.BcInfo for more details
model.bnd_rect_corner1 = [10*eps; -1];
model.bnd_rect_corner2 = [9; 2];
model.bnd_rect_index = -2;

%% fill in mandatory data in symplecicity model
% prepare matrices and vectors for computation of the Hamiltonian
model.H    = @(x,M,K,f) [K*x(1:end/2,:); ...
                         M\x(end/2+1:end,:)];
model.Hkin = @(x,M,K,f) [sparse(size(M,1), size(x,2)); ...
                         M\x(end/2+1:end,:)];
model.Hpot = @(x,M,K,f) [K*x(1:end/2,:); ...
                         sparse(size(K,1), size(x,2))];
model.h    = @(M,K,f)   [-f; zeros(size(f))];

end

% neumann values components
function res = neumann_values_components(grid, eindices, faceind, llcoord, params)
    loc = llocal2local(grid, faceind, llcoord);
    glob = local2global(grid, eindices, loc, params);

    % Set force in positive x2 direction on the right-most node layer
    res = [zeros(size(glob, 1), 1), glob(:, 1) > params.xrange(2)-10*eps];
end

% Initial values for displacement of the truncated system
function q = q0(~, model_data)
     q = zeros(model_data.truncated_df_info.ndofs, 1);
end

% Initial values for velocity of the truncated system
function v = v0(~, model_data)
    v = [0.0*model_data.truncated_df_info.grid.X; 0.0*model_data.truncated_df_info.grid.X];
end