function model = beam_non_autonomous_model(input_params)
% A realization of dyn_lin_elast_model and quadratic_symplectic_model
%   with a (rectangular) beam geometry and a time-dependent boundary
%   traction on the far-right vertical boundary
%
%   As the boundary traction is time-dependent, the model is
%   non-autonomous
%
% Patrick Buchfink, 2018

%% preallocate input parameters
if nargin < 1
    input_params = struct();
end

%% construct an autonomous beam model
model = beam_autonomous_model(input_params);

%% changes in the dynamical, linear elastic model
% add time-dependence in the Neumann values coefficients
%   this structure recycles the neumann_values_coefficients of the
%   autonomous beam model to ensure, that both models have the same
%   magnitude.
%   Therefore it is important to note the difference of params and model:
%       - model is the current beam autonomous model
%       - params is the model during runtime which has (after the following
%       line) overwritten, now time-depenent neumann_values_coefficients
model.neumann_values_coefficients = {@(params) ...
    model.neumann_values_coefficients{1}(params) * beam_model_neumann_magnitude(params)};
% update coefficients
model = model.update_coefficients(model);

%% changes in the non-autonomous framework
% change to a time-depenendent vector in the Hamiltonian
model.h = @(t,M,K,f) [-f(t); zeros(size(f(t)))];

%% change flags of time-dependent RHS coefficients
model.has_time_dependent_rhs_coefficients = true;
model.is_h_time_dependent                 = true;
% set time dependence of the Dirichlet coefficients to false
model.has_time_dependent_Dirichlet_coefficients = false;

end

% define time-dependent magnitude
function res = beam_model_neumann_magnitude(model)

    if (isfield(model, 't')) && (~isempty(model.t))
        res = 1 + 1/2*sin(2*pi*model.t/diff(model.tspan));
    else
        res = 0;
    end

end