function model = dimless_dyn_lin_elast_model()
% An abstract extension to the dyn_lin_elast_model for a dimensionless
% formulation
%
% For the derivation of the characteristic constants look in the master
% thesis P. Buchfink, Structure-preserving Model Reduction for Elasticity
%
% All parameters with an empty value are supposed to be specified by a
% "realization" of this model. Additionally, the functions with an
% "Not implemented"-error have to be specified.
%
% Patrick Buchfink, 2018

%% construct general dynamic linear elasticity model
model = dyn_lin_elast_model();

%% define characteristic constants for non-dimensionalization
% characteristic constant for lame mu
model.char_lame_mu = [];
% characteristic constant for density
model.char_density = [];
% characteristic constant for gravitational acceleration
model.char_g = [];
% characteristic constant for space variable xi to 1
%   values other than 1 would require to rescale the spatial domain omega
%   which implies that the grid in model_data needs to be recomputed (which
%   is yet not implemented)
model.char_xi = 1;

% change parameters to dimensioned quantities
model.mu_names = {'dim_lame_lambda', 'dim_lame_mu', 'dim_density'};
% redirect set_mu function
model.set_mu    = @(params, mu)    set_mu_nondim(model, params, mu);
% redirect update_coefficients function
model.apply_nondimensionalization = @apply_nondimensionalization;
% set function to set dimension time
model.set_tspan = @(params, tspan) set_tspan(params, tspan);
model.set_dt    = @(params, dt) set_dt(params, dt);

end

% apply nondimensionalization
%   use this after spezifying tspan, dt, Neumann, source and Dirichlet
%   coefficients, q0, v0, char_lame_mu, char_density, char_g
function model = apply_nondimensionalization(model)

    if isfield(model, 'is_nondim') && model.is_nondim
        error('Nondimensionalization can only be applied once!');
    end
    
    % sanity check
    if model.char_xi ~= 1
        error('char_xi ~= 1 not implemented!');
    end
    if isempty(model.char_lame_mu) || isempty(model.char_density)
        error('specify char_lame_mu and char_density before calling apply_nondimensionalization!');
    end
    
    %% calculate (dependent) characteristic constants
    % characteristic constant for time t
    model.char_t = model.char_xi * sqrt(model.char_density / model.char_lame_mu);
    % characteristic constant for displacement u
    model.char_u = (model.char_xi^2 * model.char_g * model.char_density)/ (model.char_lame_mu);
    % characteristic constant for stresses sigma
    model.char_sigma = (model.char_u*model.char_lame_mu)/model.char_xi;
    % characteristic constant for the Hamiltonian
    model.char_Ham = model.char_u^2 * model.char_lame_mu;
    % characteristic constant for velocity
    model.char_v = model.char_u/model.char_t;
    % characteristic constant for linear momentum
    model.char_p = model.char_density * model.char_v;
    
    %% switch to dimensionless formulation
    % rescale coefficients
    model = rescale_coefficients(model, 'neumann_values_coefficients', model.char_sigma);
    model = rescale_coefficients(model, 'dirichlet_values_coefficients', model.char_u);
    model = rescale_coefficients(model, 'source_coefficients', model.char_g);
    % rescale inital values
    model.q0 = @(params, model_data) model.q0(params, model_data) / params.char_u;
    model.v0 = @(params, model_data) model.v0(params, model_data) / params.char_v;
    model.p0 = @(params, model_data) model.p0(params, model_data) / params.density;
    % rescale mu, tspan and dt if necessary
    old_mu = model.get_mu(model);
    model  = model.set_mu(model, old_mu);
    model  = set_tspan(model, model.tspan);
    model  = set_dt(model, model.dt);
    % reparametrize assemble_time_dependent_force with dimensioned time t
    model.assemble_time_dependent_force = @(params, model_data, t) ...
        model.assemble_time_dependent_force(reset_temporal_data(params), model_data, params.char_t*t);
    % reparametrize assemble_time_dependent_h_or_b with dimensioned time t
    model.assemble_time_dependent_h_or_b = @(params, reduced_data, hN_or_bN_comp, t) ...
        model.assemble_time_dependent_h_or_b(rescale_tspan(params), reduced_data, hN_or_bN_comp, model.char_t*t);
    
    % set is_nondim flag
    model.is_nondim = true;
    
end

function model = rescale_coefficients(model, field, char_val)
    n_coeffs = numel(model.(field));
    coeffs   = cell(1,n_coeffs);
    for i_c=1:n_coeffs
        coeffs{i_c} = @(params) model.(field){i_c}(params) / char_val;
    end
    model.(field) = coeffs;
end

% redefined set mu
function params = set_mu_nondim(model, params, dim_mu)
    % sanity check
    if nargout == 0
      warning('set_mu_nondim called without return argument!');
    end
    
    % call default set mu default
    params = set_mu_default(params, dim_mu);
    
    % calculate rescaled, nondimensional parameters
    params.lame_lambda = params.dim_lame_lambda / params.char_lame_mu;
    params.lame_mu     = params.dim_lame_mu     / params.char_lame_mu;
    params.density     = params.dim_density     / params.char_density;
    
    % call rescale_coefficients of dyn_lin_elast_model 
    params = model.set_mu(params, dim_mu);
    
end

% set tspan to dimensionless counterpart
function params = set_tspan(params, dim_tspan)
    if nargout == 0
      warning('set_tspan called without return argument!');
    end
    params.dim_tspan = dim_tspan;
    params.tspan     = dim_tspan / params.char_t;
end

% set dt to dimensionless counterpart
function params = set_dt(params, dim_dt)
    if nargout == 0
      warning('set_dt called without return argument!');
    end
    params.dim_dt = dim_dt;
    params.dt     = dim_dt / params.char_t;
end

% reset temportal data (tsnap, dt) to dimensioned counterparts
function params = reset_temporal_data(params)
    params.dt    = params.dim_dt;
    params.tspan = params.dim_tspan;
    if isfield(params, 't')
        params.t = params.t * params.char_t;
    end
end

% rescale tspan
function params = rescale_tspan(params)
    params.tspan = params.tspan * params.char_t;
end