function model = dyn_lin_elast_model()
% An abstract model for a time-dependent linear elasticity model in rbMatlab
%   based on the original beam model
%
% All parameters with an empty value are supposed to be specified by a
% "realization" of this model. Additionally, the functions with an
% "Not implemented"-error have to be specified.
%
% Patrick Buchfink, 2018

%% problem type
% dynamical linear elasticity model
model.rb_problem_type = 'dyn_lin_elast';

% verbosity level
model.verbosity = 2;

%% parameters of the model
% names of the parameters mu
%   lame_lambda and lame_mu
%       are accessed in Fem.IntegralKernels.lin_elast_stiffness_matrix
%   density
%       is accessed in model.get_mass_matrix(...)
model.mu_names = {'lame_lambda', 'lame_mu', 'density'};

% range of the parameters mu
model.mu_ranges = {[], [], []};

% set setter and getter function for mu
model.set_mu = @set_mu_default;
model.get_mu = @get_mu_default;

%% define how the system matrix and the rhs compose
% set flags for components
model.has_volume_integral_matrix   = true; % stiffness matrix
model.has_boundary_integral_matrix = false;
model.has_volume_integral_rhs      = true; % source term (body force)
model.has_boundary_integral_rhs    = true; % Neumann b.c. (boundary traction)
model.has_dirichlet_values         = true; % Dirichl b.c.

% matrix volume integral kernel and coefficients
%   i.e. stiffness matrix
model.matrix_volume_int_kernel = ...
    {@Fem.IntegralKernels.lin_elast_stiffness_matrix};
model.matrix_volume_coeffs     = ...
    {@(params)[params.lame_mu; params.lame_lambda]};

% RHS volume integral kernel and coefficients
%   i.e. source term (body force)
model.rhs_volume_int_kernel = {@Fem.IntegralKernels.rhs_source};
% model.rhs_volume_coeffs are automatically set by model.update_coefficients

% RHS boundary integral kernel and coefficients
%   i.e. Neumann boundary condition (boundary traction)
model.rhs_boundary_int_kernel = {@Fem.IntegralKernels.rhs_neumann};
% model.rhs_boundary_coeffs are automatically set by model.update_coefficients

% compose Neumann values from coefficients and components
model.neumann_values = @(grid, eindices, faceind, llcoord, params) ...
    eval_affine_decomp_general( ...
        @(params) params.neumann_values_components(grid, eindices, faceind, llcoord, params), ...
        @(params) wrap_coeffs_to_mat_form(params, 'neumann_values_coefficients'), ...
        params);

% compose Dirichlet values from coefficients and components
model.dirichlet_values = @(grid, eindices, faceind, llcoord, params) ...
    eval_affine_decomp_general( ...
        @(params) params.dirichlet_values_components(grid, eindices, faceind, llcoord, params), ...
        @(params) wrap_coeffs_to_mat_form(params, 'dirichlet_values_coefficients'), ...
        params);

% compose the source term from coefficients and components
model.source = @(grid, eindices, loc, params) ...
    eval_affine_decomp_general(...
        @(params) params.source_components(grid, eindices, loc, params),...
        @(params) wrap_coeffs_to_mat_form(params, 'source_coefficients'), ...
        params);

% compute mass matrix from model.density
model.get_mass_matrix = @(params, model_data) ...
    params.density * model_data.truncated_df_info.l2_inner_product_matrix;

%% Dirichlet, Neumann, source terms to implement in a realization of this model
model.neumann_values_components     = ...
    @(grid, eindices, faceind, llcoord, params) error('Not implemented');
model.neumann_values_coefficients   = ...
    @(params) error('Not implemented');
model.source_components             = ...
    @(grid, eindices, loc, params) error('Not implemented');
model.source_coefficients           = ...
    @(params) error('Not implemented');
model.dirichlet_values_components   = ...
    @(grid, eindices, faceind, llcoord, params) error('Not implemented');
model.dirichlet_values_coefficients = ...
    @(params) error('Not implemented');

% second-order time-derivative of the dirichlet_value_coeficients
%   needs to be implemented for time-dependent Dirichlet values
%   and is zero for time-independent Dirichlet values
model.ddt_dirichlet_values_coefficients = @(params) 0;

%% geometry
% choose a grid generation technique
%   'construct_grid' construct a grid with the rbMatlab interal function
%                    construct_grid 
%   'pvt'            construct a grid from a pvt formulation
model.grid_generation = [];

%% time integration parameters
% flag whether the system has time-depenendet coefficients in the RHS
%   time dependence is supposed to be only in the coefficients, not the
%   components!
model.has_time_dependent_rhs_coefficients = [];

% flag whether the system has time-dependent Dirichlet coefficients
model.has_time_dependent_Dirichlet_coefficients = [];

% time span
model.tspan = [];

% time step
model.dt = [];

% time integration scheme
%   'rescaled_implicit_midpoint' implicit midpoint scheme solved fully implicitly (lin elasticity only)
%                                without direct usage of the fully inverted mass matrix
%   'rescaled_implicit_euler'    implicit euler scheme solved fully implicitly (lin elasticity only)
%                                without direct usage of the fully inverted mass matrix
model.integrator = [];

% ODE function
model.dxdt = @(params, t, x, M, K, f) ...
    dxdt(params, t, x, M, K, f);

% initial value
model.x0 = @(params, model_data) ...
    [params.q0(params, model_data); ...
     params.get_mass_matrix(params, model_data) * params.v0(params, model_data)];

% initial value for displacement
model.q0 = @(params, model_data) error('Not implemented!');
% initial value for velocity
model.v0 = @(params, model_data) error('Not implemented!');

% function to assemble the force f in the time-dependent case
model.assemble_time_dependent_force = @(params, model_data, t) ...
    assemble_time_dependent_force(params, model_data, t);

% function to assemble the time dependent bN or hN in the reduced simulation
model.assemble_time_dependent_h_or_b = @(params, reduced_data, hN_or_bN_comp, t) ...
    assemble_time_dependent_h_or_b(params, reduced_data, hN_or_bN_comp, t);

%% Model Order Reduction parameters
% list with training samples for mu
model.RB_mu_list = [];

% array with number of samples per parameter
%   number of samples = prod(n_samples_per_param)
model.n_samples_per_param = [];

% basis generation technique
%   'POD_full_state'       proper orthogonal decomposition with
%                          SVD of full state
%   'POD_separate_states'  proper orthogonal decomposition with
%                          SVD of position and linear momentum separate
%   'Lagrangian'           Lagrangian
%   'PSD_greedy'           proper symplectic decomposition, greedy algorithm (see Maboudi Afkham and Hesthaven 2017)
%   'PSD_cotangent_lift'   proper symplectic decomposition, cotangent lift (see Peng & Mohseni 2016)
%   'PSD_complex_SVD'      proper symplectic decomposition, complex SVD (see Peng & Mohseni 2016)
%   'PSD_SVD_like_decomp'  proper symplectic decomposition, SVD-like decomposition (see Buchfink, Bhatt, Haasdonk 2019)
%   'Jorthonormalized_POD' J-orthonormalized POD (see symplectic MOR paper Buchfink)
model.MOR_technique = [];

% set size of the RB
model.RB_size = [];

%% algorithmic parameters
% parameters for the FEM operators
model.operators = @Fem.Assembly.operator;
model.qdeg      = 1;
model.pdeg      = 1;

% disable caching flag. Copied from stokes_model_default.m
model.disable_caching = true;

% disable non-linearity
model.has_nonlinearity = false;

% verbosity level
model.verbose = 3;

% variable for the time which is only used in the case
%   has_time_dependent_rhs_coefficients = true
% this variable does not need to be implemented by the user, but is used by
% the system
model.t = [];

% flag for decomposition mode
%   it is set properly if needed. It's just an initialization here.
%   0: sum over coefficients * componennts
%   1: only componennts
%   2: only coefficients
model.decomp_mode = 0;

% function to update the coefficients
model.update_coefficients = @(params) update_coefficients(params);

%% function handles for the rbMatlab workflow
model.gen_model_data      = @dyn_lin_elast_gen_model_data;
model.detailed_simulation = @dyn_lin_elast_detailed_simulation;
model.plot_sim_data       = @dyn_lin_elast_plot_sim_data;
model.gen_detailed_data   = @dyn_lin_elast_gen_detailed_data;
model.gen_reduced_data    = @dyn_lin_elast_gen_reduced_data;
model.rb_simulation       = @dyn_lin_elast_reduced_simulation;
model.rb_reconstruction   = @dyn_lin_elast_rb_reconstruction;
model.get_rb_size         = @(model,detailed_data) size(detailed_data.RB, 2);
model.set_mu              = @(params, mu) set_mu(params, mu);
model.create_mu_list      = @(params, n_samples_per_param) ...
    create_mu_list(params, n_samples_per_param);

end

% ODE function
function dxdt = dxdt(params, t, x, M, K, f)
    if params.has_time_dependent_rhs_coefficients
        dxdt = [M\x(end/2+1:end); ...
                -K*x(1:end/2) + f(t)];
    else
        dxdt = [M\x(end/2+1:end); ...
                -K*x(1:end/2) + f];
    end
end
  
% function to assemble the force f in the time-dependent case
function ft = assemble_time_dependent_force(params, model_data, t)
    % save time in the model
    params.t = t;
    % compute time dependent RHS
    params.decomp_mode = 0;
    [~, ft]            = model_data.operators(params, model_data);
    
    % add inertial term of Dirichlet DOFs
    %   M_{fc} * d^2/dt^2 q^c(t)
    % the term K_{fc} * q^c(t) is automatically added by operators and thus
    %   already included in ft
    if params.has_time_dependent_Dirichlet_coefficients
        
        ddt_r_dir_coeff = wrap_coeffs_to_mat_form(params, 'ddt_dirichlet_values_coefficients');

        Mfc_ddt_r_dir = params.density * lincomb_sequence(...
            model_data.Mfc_comp, ...
            ddt_r_dir_coeff);
        
        ft = ft - Mfc_ddt_r_dir;
    end
end

% function to assemble time dependent bN or hN in the reduced simulation
function hN = assemble_time_dependent_h_or_b(model, reduced_data, hN_or_bN_comp, t)
% set current time  
model.t = t;
% get time-dependent parametric coefficients
model.decomp_mode = 2;
[~, f_coeff]      = model.operators(model, reduced_data);
% compute time dependent force
hN = lincomb_sequence(hN_or_bN_comp, f_coeff);
end

% function to create all combinations of different mus
function params = create_mu_list(params, n_samples_per_param)

    % Sanity check
    n_mus = length(params.mu_names);
    if numel(n_samples_per_param)~=n_mus
        error('Length of mu_names (%d) and n_samples_per_param (%d) differs!',...
            numel(n_samples_per_param), n_mus);
    end

    % Create samples for single parameters in mu
    %   n_samples_per_param per single parameter in mu
    mu_samples = cell(1,n_mus);
    for k=1:n_mus
        mu_samples{k} = linspace(...
            params.mu_ranges{k}(1), params.mu_ranges{k}(2), n_samples_per_param(k));
    end

    % Create all combinations of samples of single parameters in mu
    %   results in (n_samples_per_param ^ (number of parameters in mu)) samples
    mu_samples = createCombinations(mu_samples);

    % Convert combinations to RB_mu_list
    n_samples  = size(mu_samples,2);
    RB_mu_list = cell(1, n_samples);
    for k=1:n_samples
        RB_mu_list{k} = mu_samples(:,k)';
    end

    params.RB_mu_list          = RB_mu_list;
    params.n_samples_per_param = n_samples_per_param;
end

% wraps the cofficients stored in >field< in >model< from a cell formulation
%   {@(model) f1(model), @(model) f2(model), ...}
% to a matrix formulation
%   @(model) [f1(model), f2(model), ...]
function mat_form = wrap_coeffs_to_mat_form(model, field)
    num_coeffs = numel(model.(field));
    try
        mat_form   = arrayfun(@(n) model.(field){n}(model), ...
            1:num_coeffs);
    catch
        error(['Error while casting model.%s from cell to mat form! ', ...
            'Check if any model.%s yields an empty value [].'], field, field);
    end
end

% the current FEM framework needs two different representations of the
% coefficients in different cases of decomp_mode. This method allocates the
% coefficients the right way
function model = update_coefficients(model)
    model.rhs_boundary_coeffs = model.neumann_values_coefficients;
    model.rhs_volume_coeffs   = model.source_coefficients;
end

% function to set mu
function params = set_mu(params, mu)
    % sanity check
    if nargout == 0
      warning('set_mu called without return argument!');
    end
    
    % call default set mu default
    params = set_mu_default(params, mu);

    % check for phyiscal reasonfulness of parameters    
    if (params.lame_mu <= 0) || (3*params.lame_lambda + 2*params.lame_mu <= 0)
        error('parameters violate conditions for unique solution!');
    end
    if params.density <= 0
        error('density is non-positive!');
    end

    % check if mu is in range
    n_mu = numel(params.mu_names);
    for i_mu=1:n_mu
        c_name  = params.mu_names{i_mu};
        c_param = params.(c_name);
        c_range = params.mu_ranges{i_mu};
        if (c_param > c_range(2)) || (c_param < c_range(1))
            warning('The parameter ''%s'' is not in the appropraiate mu range.', c_name);
        end
    end
end

