function model = quadratic_symplectic_model(model)
% An abstract model for a symplectic model in rbMatlab
%   with a quadratic Hamiltonian with the matrix H and vector h
%
% A quadratic Hamiltonian model is a special case in this model
%   model.is_h_time_dependent = false
%
% All parameters with an empty value are supposed to be specified by a
% "realization" of this model. Additionally, the functions with an
% "Not implemented"-error have to be specified.
%
% Patrick Buchfink, 2018

% this makes the model "inheritable"
if nargin < 1
    model = struct();
end

%% function handles to evaluate Hamiltonian
model.eval_Hamiltonian = @(params, sim_data, varargin) ...
    eval_Hamiltonian(params, sim_data, varargin{:});

% cannonical Poisson matrix
model.J = @(n) [sparse(n, n), speye(n, n); -speye(n, n), sparse(n, n)];

%% symplectic parameters to be specified
% flag that shows that the model is symplectic
model.is_symplectic = true;

% matrix for the quadratic part of the Hamiltonian
model.H = @(x, varargin) error('Not implemented!');

% flag whether the vector h is time-dependent
model.is_h_time_dependent = [];

% vector for the linear part of the Hamiltonian
if model.is_h_time_dependent
    model.h = @(t, varargin) error('Not implemented!');
else
    model.h = @(varargin) error('Not implemented!');
end

end

% Compute Hamiltonian from sim_data.uh and sim_data.t
%   if the force vector is constant, the system is Hamiltonian
% Output:
%   sim_data with with Hamiltonian in sim_data.Ham for every discrete time
%   sim_data.t
function sim_data = eval_Hamiltonian(model, sim_data, varargin)

    if model.verbose > 2
        fprintf('computing Hamiltonian and its parts ...\n');
    end

    t  = sim_data.t;
    nt = numel(t);
    x  = sim_data.uh.dofs;

    if model.is_h_time_dependent
    	hx = arrayfun(@(it) dot(model.h(t(it), varargin{:}), x(:,it)), 1:nt);
    else
        hx = (model.h(varargin{:}))'*x;
    end
    
    Hx = model.H(x, varargin{:});
    % save Hamiltonian in sim_data
    sim_data.Ham = arrayfun(@(it) ...
        1/2 * dot(Hx(:,it),x(:,it)), 1:nt) ...
        + hx;

    Hx = model.Hkin(x, varargin{:});
    % save kinetic part
    sim_data.HamKin = arrayfun(@(it) ...
        full(1/2 * dot(Hx(:,it),x(:,it))), 1:nt);

    Hx = model.Hpot(x, varargin{:});
    % save potential part
    sim_data.HamPot = arrayfun(@(it) ...
        full(1/2 * dot(Hx(:,it),x(:,it))), 1:nt) ...
        + hx;
end
