function sim_data = dyn_lin_elast_detailed_simulation(model, model_data)
%function sim_data = dyn_lin_elast_detailed_simulation(model, model_data)
% Performs a detailed simulation of a dynamical, linear elastic model
%
% I. Maier, 22.01.2014
% A. Bhatt, 02.05.2017
% Patrick Buchfink, 2018

% default initializations
if ~isfield(model, 'is_symplecitc')
    model.is_symplecitc = false;
end

% initialization
start_total = tic;

% assemble truncated linear terms from model and model_data
model.decomp_mode = 0;
[K, f]            = model_data.operators(model, model_data);

% get mass matrix and inverse mass matrix
M = model.get_mass_matrix(model, model_data);

% truncated solution variable (free nodes only)
uh = Fem.DiscFunc([], model_data.truncated_df_info);

% prepare time dependency
if model.has_time_dependent_rhs_coefficients
    
    % Construct time-dependent force
    f = @(t) model.assemble_time_dependent_force(model, model_data, t);
end

% define system
odefun = @(t,x) model.dxdt(model, t, x, M, K, f);

% allocate initial state
x0 = model.x0(model, model_data);

% use a numerical method to solve the ODE
switch model.integrator
    case 'rescaled_implicit_midpoint'
        [t,dofs] = rescaled_implicit_midpoint(odefun, model.tspan, model.dt, x0, M, K);
    case 'rescaled_implicit_euler'
        [t,dofs] = rescaled_implicit_euler(odefun, model.tspan, model.dt, x0, M, K);
    otherwise
        error('Unknown numerical integrator: %s', model.integrator);
end

% save dofs in discrete function
uh.dofs     = dofs';
sim_data.uh = uh;
sim_data.t  = t;

% compute Hamiltonian with its kin. and pot part if the model is symplectic
if model.is_symplectic
    sim_data = model.eval_Hamiltonian(model, sim_data, M, K, f);
end

% compute gradproduct with the Hamiltonian matrix H if required
%   which is required for the computation of a basis for a port-Ham. sys.
%   and acts there as gradient of the Hamiltonian
if isfield(model, 'compute_Ham_product') && model.compute_Ham_product
    fprintf('Computing the product with the Hamiltonian matrix H ...\n');
    sim_data.Ham_product = model.H(dofs',M,K,f);
end

% save total time
sim_data.total_time = toc(start_total);
end