function detailed_data =  dyn_lin_elast_gen_detailed_data(model,model_data)
% Reduced basis generation with different MOR techniques
%  Lagrangian, POD_full_state, POD_separate_states, 
%
% A. Bhatt 27.4.2017
% Patrick Buchfink, April 2018

tic;

if ~isfield(model,'MOR_technique')
    model.MOR_technique = 'POD_full_state';
end

if ~isfield(model,'RB_size')
    model.RB_size = 12;
end

detailed_data = model_data;

switch(model.MOR_technique)
    case 'Lagrangian'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        % Compose ROB
        detailed_data.RB = all_dofs;
    case 'POD_full_state'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        
        % compute or load full SVD decomposition of all_dofs
        if isfield(model, 'pth_full_svd')
            if exist(model.pth_full_svd, 'file') == 2
                %load full svd
                load(model.pth_full_svd, 'U_full');
            else
                % Compute SVD
                [U_full,Sigma,~] = svd(all_dofs, 'econ');

                % compute singular values
                sing_val = diag(Sigma);

                save(model.pth_full_svd, 'U_full', 'sing_val');
            end
        end
        
        % Choose size of RB based on RB_size
        RB = U_full(:, 1:model.RB_size);
        detailed_data.RB = RB;
    case 'POD_separate_states'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        
        % compute or load separate SVD decomposition of all_dofs
        if isfield(model, 'pth_sep_svd')
            if exist(model.pth_sep_svd, 'file') == 2
                %load full svd
                load(model.pth_sep_svd, 'U_p', 'U_q');
            else
                % Compute SVD
                [U_q,~,~] = svd(all_dofs(1:end/2,:), 'econ');
                [U_p,~,~] = svd(all_dofs(end/2+1:end,:), 'econ');

                save(model.pth_sep_svd, 'U_q', 'U_p');
            end
        end
        
        % Choose size of RB based on RB_size
        k = floor(model.RB_size/2);
        k = min(k, size(U_q,2));
        detailed_data.RB = blkdiag(U_q(:,1:k), U_p(:,1:k));
    case 'PSD_greedy'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        % Compute ROB
        detailed_data.RB = psd_greedy_with_all_dofs(model, detailed_data,  all_dofs, 2);
    case 'PSD_cotangent_lift'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        % Compute ROB
        [detailed_data.RB, detailed_data.siRB] = psd_cotangent_lift(model, detailed_data,  all_dofs);
    case 'PSD_complex_SVD'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        % Compute ROB
        [detailed_data.RB, detailed_data.siRB] = psd_complex_svd(model, detailed_data,  all_dofs);
    case 'Jorthonormalized_POD'
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        detailed_data.RB = Jorthonormalized_POD(model, detailed_data,  all_dofs);
    case 'PSD_SVD_like_decomp'
        % Create all snapshots for all parameters mu in RB_mu_list
        all_dofs = run_all_RB_mu_list(model, detailed_data);
        % Compute ROB and projection matrix
        [detailed_data.RB, detailed_data.W] = psd_SVD_like_decomp(model, detailed_data,  all_dofs);
    otherwise
        error('Unknown reduction method: %s', model.MOR_technique);
end

% compute symplectic inverse for symplectic MOR techniques
MOR_tech = strsplit(model.MOR_technique, '_');
MOR_tech = MOR_tech{1};
switch(MOR_tech)
    case 'POD'
        % choose projection matrix
        detailed_data.W = detailed_data.RB;
        
    case {'PSD', 'Jorthonormalized'}
        V = detailed_data.RB;
        k = size(V,2)/2;
        n = size(V,1)/2;
        siV = model.J(k)' * V' * model.J(n);

        % choose projection matrix to be the transposed of the symplectic inverse
        detailed_data.W = siV';

        % Check if model is symplectic
        if ~isfield(model, 'is_symplectic') || ~model.is_symplectic
            warning('Model does not have the is_symplectic flag, but a symplectic MOR technique is used');
        end

        % Error in Hamiltonian
        x0 = model.x0(model,detailed_data);
        [delta_ham, rel_deltaHam] = ...
            compute_deltaHam_for_RB_mu_list(model, detailed_data, x0, V*siV*x0);
        fprintf('max. abs. difference in Hamiltonian at t=0: %4.2e\n', max(delta_ham));
        fprintf('max. rel. difference in Hamiltonian at t=0: %4.2e\n', max(rel_deltaHam));
        
    otherwise
        error('Unknown MOR technique: %s', MOR_tech);
end

% Sanity check for RB size
if size(detailed_data.RB,2) ~= model.RB_size
    errorStruct.message = sprintf('The size of the given reduced basis is %d and does not equal the preferred size %d!', ...
        size(detailed_data.RB,2), model.RB_size);
    errorStruct.identifier = 'RBmatlab:DynLinElast:GenDetailedData:RBSizeMissmatch';
    errorStruct.actual_RB_size = size(detailed_data.RB,2);
    
    detailed_data.actual_RB_size = size(detailed_data.RB,2);
    
    error(errorStruct);
else
    detailed_data.actual_RB_size = model.RB_size;
end

end

function [all_dofs, all_Ham_product] = run_all_RB_mu_list(model, detailed_data)

n_RB_mu_list = numel(model.RB_mu_list);

% path to save snapshots
if isfield(model, 'pth_snapshots')
    pth_snapshots = model.pth_snapshots;
else
    pth_snapshots = '';
end

all_dofs = [];
all_Ham_product = [];
all_max_state = [];

% enable computation of the production with the Hamiltonian matrix
model.compute_Ham_product = true;

if exist(pth_snapshots,'file') == 2
    % Load model
    load(pth_snapshots, 'all_dofs', 'all_Ham_product');
else
    for i_RB_mu_list = 1:n_RB_mu_list
        fprintf('running detailed simulation %d of %d ...\n', i_RB_mu_list, n_RB_mu_list);

        % Set current mu
        model = set_mu(model, model.RB_mu_list{i_RB_mu_list});

        % Run next simulation
        sim_data = detailed_simulation(model,detailed_data);
        
        % Save simulation dofs
        cur_dofs = sim_data.uh.dofs;
        all_dofs = [all_dofs, cur_dofs];
        all_Ham_product   = [all_Ham_product, sim_data.Ham_product];
        max_norm_cur_dofs = sqrt(max(sum(cur_dofs.^2, 1)));
        all_max_state = [all_max_state; max_norm_cur_dofs * ones(size(cur_dofs,2),1)];
    end

    if ~isempty(pth_snapshots)
        save(pth_snapshots, 'all_dofs', 'all_Ham_product', 'all_max_state');
    end
end

fprintf('Finished with all %d simulations.\n', n_RB_mu_list);

end

% Maboudi, Algorithm 2 (with modified initial step)
function [A, siA] = psd_greedy(model, detailed_data, x0)

%%% Initializations
n = size(x0,1)/2;

% Symplectic matrix
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];

% Dirty trick for initial step
%   Basically it executes the while loop below with idx_mu = 1
delta_ham = Inf;
rel_deltaHam = Inf;
e = zeros(2*n,1);
f = zeros(2*n,1);
A = [e,f];
k = 0;
siA = A';

% Parameter
delta = 1e-5;

% Cell array containing snapshots
n_RB_mu_list = numel(model.RB_mu_list);
S = cell(1, n_RB_mu_list);

while any(rel_deltaHam > delta)
    % Find parameter with biggest difference in Hamiltonian
    [~, idx_mu] = max(delta_ham);
    
    % Compute snapshots of parameter idx_mu if neccessary
    if isempty(S{idx_mu})
        % Run simulation
        sim_data = detailed_simulation(model, detailed_data);

        % Save simulation dofs
        S{idx_mu} = sim_data.uh.dofs;
    end
    
    % Find snapshot with biggest error
    [~,idx_S] = max(sum((S{idx_mu} - A*siA*S{idx_mu}).^2,1)); %sqrt not needed due to monotonicity
    zstar = S{idx_mu}(:,idx_S);

    % Apply symplectic Gram-Schmidt on zstar
    ztilde = zstar - e * (e'*zstar) - f * (f'*zstar);
    
    % Save two new basis vectors
    e(:,k+1) = ztilde/norm(ztilde);
    f(:,k+1) = J2n'*e(:, k+1);
    A        = [e,f];

    k = k + 1;

    % Compute symplectic inverse
    J2k = [sparse(k, k), speye(k, k); ...
           -speye(k, k), sparse(k, k)];
    siA = J2k' * A' * J2n;

    % Error in Hamiltonian
    [delta_ham, rel_deltaHam] = ...
        compute_deltaHam_for_RB_mu_list(model, detailed_data, x0, A*siA*x0);
    fprintf('max. abs. difference in Hamiltonian at t=0: %4.2e\n', max(delta_ham));
    fprintf('max. rel. difference in Hamiltonian at t=0: %4.2e\n', max(rel_deltaHam));
end

warning('PSD greedy ignores RB_size in model. Resultant RB_size is: %d', size(A,2));

end

% Maboudi, Algorithm 2 (with modified initial step)
% modified such that it generates a RB with any size
% changed termination criterion in the loop and initial step
function A = psd_greedy_with_all_dofs(model, ~, Xs, lmax)

%%% Initializations
n = size(Xs,1)/2;

% Symplectic matrix
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];

% choose desired basis size
k = floor(model.RB_size/2);
k = min(k, size(Xs,2));

% Initializations
A   = zeros(2*n,2*k);
L   = zeros(2*k,2*k);
siA = A';
J2k = [sparse(k, k), speye(k, k); ...
       -speye(k, k), sparse(k, k)];

for j = 1:k
    
    % Find snapshot with biggest error
    [err,idx_S] = max(sum((Xs - A*siA*Xs).^2,1)); %sqrt not needed due to monotonicity
    if err < sqrt(eps)
        A(:,[j:k,k+j:end]) = [];
        break;
    end
    zstar = Xs(:,idx_S);
    
    % SR method
    [L, A] = inc_MSGS([zstar,J2n'*zstar], j, L, A, lmax);

    % Compute symplectic inverse
    siA = J2k' * A' * J2n;
end

end

function [A, siA] = psd_cotangent_lift(model, ~, Xs)

%%% Initializations
n = size(Xs,1)/2;

%%% Combine snapshots
S_combined = [Xs(1:n,:), Xs((n+1):end,:)];

%%% Compute left singular vectors
[U,~,~] = svd( S_combined, 'econ' );

k = floor(model.RB_size/2);
k = min(k, size(U,2));

%%% Save ROB
A = blkdiag(U(:,1:k), U(:,1:k));

%%% Compute symplectic inverse
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];
J2k = [sparse(k, k), speye(k, k); ...
           -speye(k, k), sparse(k, k)];
siA = J2k' * A' * J2n;

end

function [A, siA] = psd_complex_svd(model, ~, Xs)

%%% Initializations
n = size(Xs,1)/2;

%%% Combine snapshots
S_complex = Xs(1:n,:) + 1i*Xs((n+1):end,:);

%%% Compute left singular vectors
[U,~,~] = svd( S_complex, 'econ' );

k = floor(model.RB_size/2);
k = min(k, size(U,2));

%%% Save ROB
Phi = real(U(:, 1:k));
Psi = imag(U(:, 1:k));
A = [Phi, -Psi; ...
     Psi,  Phi];

%%% Compute symplectic inverse
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];
J2k = [sparse(k, k), speye(k, k); ...
           -speye(k, k), sparse(k, k)];
siA = J2k' * A' * J2n;

end

function [deltaHam, rel_deltaHam] = compute_deltaHam_for_RB_mu_list(model, detailed_data, x1, x2)

n_RB_mu_list = numel(model.RB_mu_list);

model.decomp_mode = 0;

% get mass matrix
M = model.get_mass_matrix(model, detailed_data);

deltaHam     = zeros(1, n_RB_mu_list);
rel_deltaHam = zeros(1, n_RB_mu_list);
for i_RB_mu_list = 1:n_RB_mu_list
    % Set current mu
    model = set_mu(model, model.RB_mu_list{i_RB_mu_list});
    
    % Get operators
    if model.has_time_dependent_rhs_coefficients
        % construct stiffness matrix
        model.decomp_mode = 0;
        K                 = detailed_data.operators(model, detailed_data);
        % Construct time-dependent force
        f = @(t) model.assemble_time_dependent_force(model, detailed_data, t);
    else
        model.decomp_mode = 0;
        [K,f]             = detailed_data.operators(model, detailed_data);
    end
    
    % Compute Hamiltonian for current parameters
    sim_data.t = model.tspan(1);
    sim_data.uh.dofs = x1;
    sim_data1 = model.eval_Hamiltonian(model, sim_data, M, K, f);
    sim_data.uh.dofs = x2;
    sim_data2 = model.eval_Hamiltonian(model, sim_data, M, K, f);
    
    % Compute difference in Hamiltonian
    deltaHam(i_RB_mu_list)     = abs(sim_data1.Ham - sim_data2.Ham);
    rel_deltaHam(i_RB_mu_list) = deltaHam(i_RB_mu_list)/abs(sim_data1.Ham);
end

end

function A = Jorthonormalized_POD(model, ~, Xs)

%%% Initializations
n = size(Xs,1)/2;

% compute or load full SVD decomposition of all_dofs
if isfield(model, 'pth_full_svd')
    if exist(model.pth_full_svd, 'file') == 2
        %load full svd
        load(model.pth_full_svd, 'U_full');
    else
        % Compute SVD
        [U_full,Sigma,~] = svd(Xs, 'econ');

        % compute singular values
        sing_val = diag(Sigma);

        save(model.pth_full_svd, 'U_full', 'sing_val');
    end
end

k = floor(model.RB_size/2);
k = min(k, size(U_full,2));

%%% Compute symplectic inverse
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];
       
%%% Save ROB
% MSGS method with skipping unappropriate pairs of vectors
A = MSGS_with_skip([U_full, J2n'*U_full], k, 2);

end

% Modified symplectic Gram–Schmidt algorithm (MSGS)
% see Salam (2005) On theoretical and numerical aspects of symplectic Gram–Schmidt-like algorithms
%   lmax = 1: MSGS
%   lmax = 2: MSGSR
function S = MSGS(B, lmax)

% Initializations
n = size(B,1)/2;
p = size(B,2)/2;
   
S = zeros(2*n, 2*p);
L = zeros(2*p, 2*p); %Lambda

for j=1:p
    Wj = B(:,[j,p+j]);
    % one step with MSGS / MSGSR
    [L,S] = inc_MSGS(Wj, j, L, S, lmax);
end

end

% Modified symplectic Gram–Schmidt algorithm (MSGS) with skip
% see Salam, 2005, On theoretical and numerical aspects of symplectic Gram–Schmidt-like algorithms
%   lmax = 1: MSGS
%   lmax = 2: MSGSR
% In addition, algorithm does not break down on non-isotropic images, but
% skips this pair of vectors
function [S, idx_max] = MSGS_with_skip(B, k, lmax, ESR_type)

if nargin < 4
    ESR_type = 2;
end

% Initializations
n     = size(B,1)/2;
max_i = size(B,2)/2;
   
S = zeros(2*n, 2*k);
L = zeros(2*k, 2*k); %Lambda

idx = 1;

for j=1:k
    % Initial value for is isotropic flag
    is_isotropic = true;
    while is_isotropic && idx<=max_i
        % presume the vectors not to be isotropic
        is_isotropic = false;
        % get new pair of vectors
        Wj = B(:,[idx,max_i+idx]);
        idx  = idx+1;
        % one step with MSGS / MSGSR
        try
            [L,S] = inc_MSGS(Wj, j, L, S, lmax, ESR_type);
        catch
            is_isotropic = true;
        end
    end
    
    if idx>max_i
        if is_isotropic
            S(:, j:end) = [];
        else
            S(:, (j+1):end) = [];
        end
        break;
    end
end

idx_max = idx;

end

% incremental step of MSGS
function [L, S] = inc_MSGS(Wj, j, L, S, lmax, ESR_type)

if nargin < 6
    ESR_type = 2;
end

% Initializations
n = size(S,1)/2;
p = size(S,2)/2;

J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];
J2  = [ 0, 1; ...
       -1, 0];

% lmax = 1: MSGS
% lmax = 2: MSGSR
for l=1:lmax
    for i=1:(j-1)
        L([i,p+i],[j,p+j]) = J2' * (S(:,[i,p+i]))' * J2n * Wj;
        Wj = Wj - S(:,[i,p+i]) * L([i,p+i],[j,p+j]);
    end
end

% check if vectors are non-isotropic
if abs(Wj(:,1)'*J2n*Wj(:,2)) < sqrt(eps)
    error('inc_MSGS:IsotropicImage', 'Image of Wj(:,1) and Wj(:,2) is isotropic!');
end

% evaluate ESR1,2 or 3 based on the ESR_type given in the parameters
[V,H] = ESR(Wj, ESR_type);
S(:,[j,p+j])       = V;
L([j,p+j],[j,p+j]) = H;
end

% Elementary SR factorization (ESR)
% see Salam, 2005, On theoretical and numerical aspects of symplectic Gram–Schmidt-like algorithms
function [V, H] = ESR(A, ESR_type)

if size(A,2) ~= 2
    error('ESR is for size(A,2) = 2 only!');
end

n = size(A,1)/2;

J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];

switch(ESR_type)
    case 1
        r11 = norm(A(:,1));
        r12 = 0; % best solution in num. experiments in Salam, 2005
    case 2 % most advanced according to Salam, 2005
        r11 = norm(A(:,1));
        v1  = A(:,1)/r11;
        r12 = v1'*A(:,2);
    case 3
        alpha = A(:,1)'*J2n*A(:,2);
        r11 = sqrt(alpha);
        r12 = 0; % best solution in num. experiments in Salam, 2005
    case 4 % Della-Dora
        r11 = 1;
        r12 = 0;
        
    otherwise
        error('Unknown ESR type: %d', ESR_type);
end

v1  = A(:,1)/r11;
w   = A(:,2) - r12 * v1;
r22 = v1'*J2n*w;
v2  = w/r22;

V = [v1, v2];
H = [r11, r12; 0, r22];

end

function [V, siV] = psd_SVD_like_decomp(model, ~,  Xs)

% init system dimension and corresponding Poisson matix
n = size(Xs, 1) / 2;
J2n = model.J(n);

% choose desired basis size
k = floor(model.RB_size/2);
k = min(k, size(Xs,2));
J2k = model.J(k);


% compute or load full SVD-like decomposition of Xs
if isfield(model, 'pth_svd_like')
    if exist(model.pth_svd_like, 'file') == 2
        %load svd-like decomposition
        load(model.pth_svd_like, 'siST', 'w_sympl_sing_val', 'p', 'q');
    else
        [S, D, ~, p, q, meta_data] = svd_like_decomposition(Xs');
    
        % compute the transposed of the symplectic inverse to switch from the
        % notation of the original paper to Xs = siST*D'*Q'
        siST = J2n'*S*J2n;
        n = size(siST,1) / 2;

        % compute weighted symplectic singular values
        sympl_sing_val   = full(diag(D));
        sympl_sing_val   = sympl_sing_val(1:p);
        col_norm_S       = sqrt(sum(S(:,1:p).^2, 1) + sum(S(:,n+(1:p)).^2, 1));
        w_sympl_sing_val = sympl_sing_val .* col_norm_S(:);
        weighted_q       = sum(S(:,p+(1:q)).^2, 1);
        w_sympl_sing_val = [w_sympl_sing_val(:); weighted_q(:)];

        save(model.pth_svd_like, 'siST', 'sympl_sing_val', 'w_sympl_sing_val', 'p', 'q', 'meta_data');
    end
end

[~, idx] = sort(w_sympl_sing_val, 'desc');
if k > p+q
    idx = [idx(:)', p+q+(1:k-p-q)];
else
    idx = idx(1:k);
end

% select the first k pairs of symplectic basis vectors
V = siST(:, [idx, n+idx]);

% compute symplectic inverse
siV = J2k' * V' * J2n;

end
