function model_data = dyn_lin_elast_gen_model_data(model)
%function model_data = dyn_lin_elast_gen_model_data(model)
% Create model data for a dynamical linear, elastic model
%   The Dirichlet values are implemented by truncataing constrainted nodes
%
% Be aware that different variables are designed for the full system and
% some for the truncated sysem.
% Full system variables:
%   df_info, bc_info
% Truncated system variables:
%   truncated_df_info, grid, operators

if ~model.has_dirichlet_values
    error('dyn_lin_elast_gen_model_data is only implemented for models with Dirichlet values');
end

%% construct full system (free + Dirichlet nodes)
% construct grid with rbMatlab methods
switch(model.grid_generation)
    case 'construct_grid'
        grid = construct_grid(model);
    case 'pet'
        [p, ~, t] = model.get_geometry_pet(model);
        % Construct grid from truncated p-e-t formulation
        grid = triagrid(p, t, model);
        grid = set_boundary_types(grid, model);
    otherwise
        error('Unknown grid generation technique: %s', model.grid_generation);
end

%% construct discrete function spaces (free + Dirichlet nodes)
scalar_df_info = Fem.Lagrange.DefaultInfo(model.pdeg, grid);
df_info = Fem.CompositeFunctionSpace(...
    scalar_df_info, 'x_comp', ...
    scalar_df_info, 'y_comp');

bc_info = Fem.BcInfo(model, grid, df_info);

model_data.df_info = df_info;
model_data.bc_info = bc_info;

%% construct truncated system (free nodes only)
truncated_scalar_df_info = ...
    Fem.Lagrange.TruncatedInfo(model, scalar_df_info, bc_info);
truncated_df_info = Fem.CompositeFunctionSpace(...
    truncated_scalar_df_info, 'x_comp', ...
    truncated_scalar_df_info, 'y_comp');

model_data.grid              = truncated_df_info.grid;
model_data.truncated_df_info = truncated_df_info;

model_data.operators = Fem.OperatorsTruncated();

%% assemble operators
if model.has_nonlinearity
    warning(['Take special care with pre-assembled systems and ', ...
        'nonlinearity! This case has not been tested.']);
end
model_data.operators.assemble_components(model, model_data);

%% compute M_{fc} times components of the Dirichlet values
%   which are used if the Dirichlet values are time-dependent
if model.has_time_dependent_Dirichlet_coefficients
    if iscell(model_data.bc_info.dirichlet_dof_vector_components)
        diri_comp = model_data.bc_info.dirichlet_dof_vector_components;
    else
        diri_comp = {model_data.bc_info.dirichlet_dof_vector_components};
    end
    
    n_diri_comp = numel(diri_comp);
    Mfc_comp    = cell(n_diri_comp,1);
    
    for ic=1:n_diri_comp
        % compute M times diri_comp, where M is the global / non-truncated mass matrix
        Mfc_comp{ic} = model_data.df_info.l2_inner_product_matrix * diri_comp{ic};
        % keep only the part of the free DOFs (i.e. M_{fc} times diri_comp)
        % by deleting the part of the dirichlet DOFs
        Mfc_comp{ic}(model_data.bc_info.dirichlet_gids) = [];
    end
    
    % save the compoted compopnents in model_data
    model_data.Mfc_comp = Mfc_comp;

end
