function reduced_data = dyn_lin_elast_gen_reduced_data(model, detailed_data)
%function reduced_data = beam_gen_reduced_data(model,detailed_data)
%
% function computing reduced data.

% A. Bhatt 27.4.2017
% Patrick Buchfink, April 2018

% Prepare FEM matrices (aka opertors)
op = detailed_data.operators;
if ~op.componentsAssembled
    op.assemble_components(model, detailed_data);
end

% Prepare mass matrix
M = model.get_mass_matrix(model, detailed_data);

% Save number of components for A and f
Q_A = length(op.A_comp);
reduced_data.Q_A = Q_A;
Q_f = length(op.f_comp);
reduced_data.Q_f = Q_f;

% Select projection matrices
V = detailed_data.RB;
W = detailed_data.W;

% Hint: Meddling with (1:end/2) and (end/2+1:end) stems from the transition
%   from a second-order system to a first-order system.
%   detailed_data.operators are valid for the second order system and
%   "thus are only half of the first order system"
% The first order system is A = [0, M^{-1}; -K, 0], b = [0; f]
% where K = op.A_comp and f = op.f_comp come from detailed_data.operators

% Compute reduced components of stiffness matrix
reduced_data.AN_comp = cell(1, Q_A);
for q = 1:Q_A
    reduced_data.AN_comp{q} = -W(end/2+1:end,:)' * op.A_comp{q} * V(1:end/2,:);
end
% Compute reduced inverse of mass matrix (which will be part of A)
reduced_data.AinvMN = W(1:end/2,:)' * (M \ V(end/2+1:end,:));

% Compute reduced components of input vector
reduced_data.bN_comp = cell(1, Q_f);
for q = 1:Q_f
    reduced_data.bN_comp{q} = W(end/2+1:end,:)' * op.f_comp{q};
end

% Compute reduced initial state
reduced_data.x0 = W' * model.x0(model, detailed_data);

reduced_data.q0 = reduced_data.x0(1:end/2);
reduced_data.p0 = reduced_data.x0(end/2+1:end);

% Comuputation of the reduced Hamiltonian
if isfield(model, 'is_symplectic') && model.is_symplectic
    % Hint: Again meddling with (1:end/2) and (end/2+1:end) due to the same
    %   reasons mentioned above
    % The Hamiltonian in the detailed model is H = blkdiag(K, M^{-1}), h = [-f; 0]

    % Compute reduced components of Hamiltonian (quadratic part with stiffness matrix)
    reduced_data.HN_comp = cell(1, Q_A);
    for q = 1:Q_A
        reduced_data.HN_comp{q} = V(1:end/2,:)' * op.A_comp{q} * V(1:end/2,:);
    end
    % Compute reduced components of Hamiltonian (quadratic part with mass matrix)
    reduced_data.HinvMN = V(end/2+1:end,:)' * (M \ V(end/2+1:end,:));

    % Compute reduced components of Hamiltonian (linear part)
    reduced_data.hN_comp = cell(1, Q_f);
    for q = 1:Q_f
        reduced_data.hN_comp{q} = V(1:end/2,:)' * (-op.f_comp{q});
    end

    % Create handle for Hamiltonian
    reduced_data.eval_reduced_Hamiltonian = @eval_reduced_Hamiltonian;
end

end

% online computation of the reduced Hamiltonian
function reduced_sim_data = eval_reduced_Hamiltonian(model, reduced_data, K_coeff, f_coeff, reduced_sim_data)

t  = reduced_sim_data.t;
xr = reduced_sim_data.reduced_dofs;

nt = numel(t);
HN = lincomb_sequence(reduced_data.HN_comp, K_coeff) + reduced_data.HinvMN;

if model.has_time_dependent_rhs_coefficients
    hN = @(t) model.assemble_time_dependent_h_or_b(model, reduced_data, reduced_data.hN_comp, t);
else
    hN = lincomb_sequence(reduced_data.hN_comp, f_coeff);
end

% time-dependent case
if model.has_time_dependent_rhs_coefficients
    hNxr = arrayfun(@(it) dot(hN(t(it)), xr(:,it)), 1:nt);

% time-independent case
else
    hNxr = hN'*xr;
end

% Hamiltonian
HNxr = HN*xr;
reduced_sim_data.Ham = arrayfun(@(n) 1/2*dot(HNxr(:,n), xr(:,n)), 1:nt) + hNxr;

% kinetic part
reduced_sim_data.HamKin = arrayfun(@(n) 1/2*dot(reduced_data.HinvMN*xr(:,n), xr(:,n)), 1:nt);

% fast computation of potential part of Hamiltonian
reduced_sim_data.HamPot = reduced_sim_data.Ham - reduced_sim_data.HamKin;

end
