function p = dyn_lin_elast_plot_sim_data(model, model_data, sim_data, plot_params)
%function p = beam_plot_sim_data(model, model_data, sim_data, plot_params)
%

if nargin < 4
    plot_params = [];
end

if ~isfield(plot_params, 'axis_equal')
    plot_params.axis_equal = 1;
end

if ~isfield(plot_params, 'title_on')
    plot_params.title_on = 1;
end

if ~isfield(plot_params, 'mode')
    plot_params.mode = {'displacement'};
end

if ~isfield(plot_params, 'displacement_scale')
    plot_params.displacement_scale = 1;
end

% Sanity checks
if isfield(plot_params, 'modes')
    error('beam_plot_sim_data does only provide the plot_params.mode field');
end

if iscell(plot_params.mode)
    error('plot_params.mode can not be a cell array');
end


% Following methods are printed with slider, if several state vectors are
% handed over to this method
if size(sim_data.uh.dofs,2)>1 && ...
   ismember(plot_params.mode, {'displacement', 'displacement_abs', 'displacement_scatter'})

    % Set title string if there is none yet
    if ~isfield(plot_params, 'title')
        switch(plot_params.mode)
            case {'displacement'}
                plot_params.title = 'displacement (with grid)';
            case {'displacement_abs'}
                plot_params.title = 'displacement (absolute values)';
            case {'displacement_scatter'}
                plot_params.title = sprintf('displacement (without grid) scaled by %4.2e', plot_params.displacement_scale);
        end
    end
    
    plot_params.plot = @(grid, data, params) ...
        plot_sequence_wrapper(grid, data, params, model, model_data, sim_data);
    p = plot_sequence(sim_data.uh.dofs, model_data.grid, plot_params);
    return;
end

% Check if this is a call from plot_sequence, if not: open a new figure
if ~isfield(plot_params, 'plot')
    f = figure;
end

switch plot_params.mode

    case {'displacement'}

        uh_x = get_component(sim_data.uh, 1);
        uh_y = get_component(sim_data.uh, 2);

        X = model_data.grid.X + uh_x.dofs*plot_params.displacement_scale;
        Y = model_data.grid.Y + uh_y.dofs*plot_params.displacement_scale;

        dis_grid = triagrid([X, Y]', model_data.grid.VI', model);

        range(1) = min([model.xrange(1); X]);
        range(2) = max([model.xrange(2); X]);
        range(3) = min([model.yrange(1); Y]);
        range(4) = max([model.yrange(2); Y]);
        
        p = plot(dis_grid);
        axis equal
        axis(range);

    case {'displacement_abs'}

        uh_x = get_component(sim_data.uh, 1);
        uh_y = get_component(sim_data.uh, 2);

        uh_abs = Fem.DiscFunc([], uh_x.df_info);
        uh_abs.dofs = sqrt(uh_x.dofs.^2 + uh_y.dofs.^2)*plot_params.displacement_scale;

        p = plot(uh_abs, plot_params);

    case {'displacement_scatter'}

        uh_x = get_component(sim_data.uh, 1);
        uh_y = get_component(sim_data.uh, 2);

        X = model_data.grid.X + uh_x.dofs*plot_params.displacement_scale;
        Y = model_data.grid.Y + uh_y.dofs*plot_params.displacement_scale;

        
        hold('on');
        % Plot original part
        scatter(model_data.grid.X, model_data.grid.Y, 'x');
        % Plot displaced part
        p = scatter(X, Y, 'o');
        hold('off');

    case {'plot_Hamiltonian'}

        p = plot(sim_data.t(:), sim_data.Ham(:), ...
                sim_data.t(:), sim_data.HamKin(:), ...
                sim_data.t(:), sim_data.HamPot(:));
        legend('Hamiltonian', 'kin. part', 'pot. part');
        
        if plot_params.title_on
            if ~isfield(plot_params, 'title')
                plot_params.title = 'Hamiltonian over time';
            end
            set(f,'Name',plot_params.title);
        end
        
    case {'plot_error_Hamiltonian'}
        
        figure();

        p = plot(sim_data.t, (sim_data.Ham - sim_data.Ham(1)));
        
        if plot_params.title_on
            if ~isfield(plot_params, 'title')
                plot_params.title = 'error in the Hamiltonian over time';
            end
            set(f,'Name',plot_params.title);
        end

    case {'error'}
        
        err = sqrt(sum((sim_data.uh.dofs - plot_params.detailed_uh.dofs).^2, 1));
        
        p = plot(sim_data.t, err);

        if plot_params.title_on
            if ~isfield(plot_params, 'title')
                plot_params.title = 'Error (unweighted) over time';
            end
            set(f,'Name',plot_params.title);
        end

    case {'error_weighted'}
        
        model.decomp_mode = 0;
        K                 = model_data.operators(model, model_data);
        M                 = model_data.M;

        H = @(x) 1/2*[K*x(1:end/2); M\x(end/2+1:end)];
        
        err = sim_data.uh.dofs - plot_params.detailed_uh.dofs;
        err = sqrt(arrayfun(@(n) err(:,n)' * H(err(:,n)), 1:length(sim_data.t)));

        if plot_params.title_on
            if ~isfield(plot_params, 'title')
                plot_params.title = 'Error (weighted with H) over time';
            end
            set(f,'Name',plot_params.title);
        end
        
        p = plot(sim_data.t, err);
        
    otherwise
        p = [];
        warning('Unknown plot mode: %s', plot_params.mode);
end

if plot_params.axis_equal
    axis equal
end

end

function plot_sequence_wrapper(grid, data, plot_params, model, model_data, sim_data)
% this is an intermediatory function to enable using the legacy plot
% function

model_data.grid = grid;
new_sim_data.uh = Fem.DiscFunc(data, sim_data.uh.df_info);
beam_plot_sim_data(model, model_data, new_sim_data, plot_params);

end