% Create bridge model
function [p,e,t] = get_geometry_pet_bridge(params)

% Length
l = params.geo_length;

% Number of sections
nsec = params.geo_num_sections;

% Thickness of lane
tl = params.geo_thickness_lane;

% Thickness of lane supporting structure
ts = params.geo_thickness_structure;

% height on outer
ho = params.geo_height_left_right;

% height in the center
hm = params.geo_height_center;

% create lane
lane = [3; 4;                                % Says: Rectangle
        l+ts/2;   -ts/2;    -ts/2;  l+ts/2;  % x values
        hm-tl/2;  hm-tl/2; hm+tl/2; hm+tl/2; % y values
        0; 0];                               % empty values
    
arc = @(x) -((x - l/2)/(l/2)).^2 * (ho - hm);

% create sections
lsec = l/nsec;
secs = cell(1, 4*nsec+1);
names = cell(1, numel(secs));
for i_sec=1:nsec
    xr = i_sec * lsec;
    xl = xr - lsec;
    xm = (xl + xr)/2;
    
    % conjunction
    i_con       = 4*i_sec;
    secs{i_con} = [3; 4;                                                          % Says: Rectangle
                  construct_arc(arc, max(xl-lsec,0), xl, xr, min(xr+lsec,l), ts); % x and y values
                  0; 0];                                                          % empty slots
    
    % extract left upper conjunction point
    ylucon = secs{i_con}(6+2);
    % extraft right upper conjuction point
    yrucon = secs{i_con}(6+3);
    % compute conjuction upper line slope
    alpha = atan((yrucon-ylucon)/(xr-xl));
    % left diagonal
    i_ldiag = 4*i_sec-2;
    secs{i_ldiag} = [3; 4;                                               % Says: Rectangle
                    construct_ldiag(xl, ylucon, xm, hm-tl/2, alpha, ts); % x and y values
                    0; 0];                                               % empty slots
    % right diagonal
    i_rdiag = 4*i_sec-1;
    secs{i_rdiag} = [3; 4;                                               % Says: Rectangle
                    construct_rdiag(xr, yrucon, xm, hm-tl/2, alpha, ts); % x and y values
                    0; 0];                                               % empty slots
    
    % left column
    i_lcol       = 4*i_sec-3;
    if i_sec > 1 && i_sec <= nsec
        rdiag = secs{i_rdiag-4};
    else
        rdiag = [];
    end
    ldiag = secs{i_ldiag};
    secs{i_lcol} = [3; 4;                             % Says: Rectangle
                    construct_lcol(ldiag, rdiag, ts); % x and y values
                    0; 0];                            % empty slots
    
    % Save names for left column, left diagonal, right diagonal and conjunction
    char_i = num2str(i_sec);
    names(1, i_lcol:i_con) = {['LC',char_i], ...
                              ['LD',char_i], ...
                              ['RD',char_i], ...
                              ['CN',char_i]};
end

% Add last right column
delta = atan((secs{i_rdiag}(2+2) - secs{i_rdiag}(2+1))/ ...
    (secs{i_rdiag}(6+2) - secs{i_rdiag}(6+1)));
h_rcol = (hm-tl/2) + (xr-secs{i_rdiag}(2+2)-ts/2)/tan(delta);
secs{end} = [3; 4;                              % Says: Rectangle
            xr-ts/2; xr-ts/2; xr+ts/2; xr+ts/2; % x values
            hm-tl/2; h_rcol;  h_rcol;  hm-tl/2; % y values
            0; 0];                              % empty slots
            
names(end) = {'RC'};

% Add lane to names
names = {'LANE', names{:}};

% Add triangle fill to inner conjunctions
fc_names = cell(1,nsec+1);
fcs      = cell(1,nsec+1);
for i_fc = 2:nsec
    % Extract coordinates to fill from secs
    lcon = secs{4*i_fc};
    lcol = secs{4*i_fc-3};
    
    fcs{i_fc-1} = [2; 3;                           % Says: Triangle
                  lcon(2+2); lcol(2+2); lcol(2+3); % x values
                  lcon(6+2); lcol(6+2); lcol(6+3); % y values
                  0; 0; 0; 0];                     % empty rows
    % save name
    fc_names{i_fc-1} = ['FC',num2str(i_fc)];
end

% Add pentagon fill to left conjunction
fc_names{end-1} = 'FCL';
lcol = secs{1};
lcon = secs{4};

fcs{end-1} = [2; 5;                                                       % Says: Pentagon
             lcon(2+2); lcol(2+3); lcol(2+2);  lcol(2+2); lcon(2+1);  ... % x values
             lcon(6+2); lcol(6+3); lcol(6+2);  lcon(6+2); lcon(6+1)]; ... % y values
             

% Add pentagon fill to right conjunction
fc_names{end} = 'FCR';
rcon = secs{end-1};
rcol = secs{end};

fcs{end} = [2; 5;                                                       % Says: Pentagon
            rcon(2+3); rcol(2+2); rcol(2+3); rcol(2+3); rcon(2+4);  ... % x values
            rcon(6+3); rcol(6+2); rcol(6+3); rcon(6+3); rcon(6+4)]; ... % y values

names = {names{:}, fc_names{:}};

% Create geometry
gd = [lane, secs{:}, fcs{:}];
ns = char(names{:})';
sf = strjoin(names(1:end), '+');
g = decsg(gd,sf,ns);

% pdegplot(g,'EdgeLabels','off','SubdomainLabels','off');
% axis equal;

% create mesh
pdem = createpde();
geometryFromEdges(pdem, g);
generateMesh(pdem);

% % plot mesh
% pdeplot(pdem);
% axis equal;

% obtain pet formulation
[p,e,t] = meshToPet(pdem.Mesh);

% translate all nodes by ts/2 to avoid negative values for nodes
%   in x and y direction!
%   this is just a cosmetic thing
p = p + ts/2;

end

% x and y coordinates of the centerline (lower and upper) and thickness t
function coords = construct_arc(arc, x1, x2, x3, x4, t)

% calculate translation
[ytl, ytr] = mean_arc_height(arc, x1, x2, x3, x4, t);

% compute arc
y2 = arc(x2);
y3 = arc(x3);

% calculate coordinates with translation
x = [x2; x2; x3; x3];
y = [y2 - ytl; y2 + ytl; y3 + ytr; y3 - ytr];

% return x and y coordinates
coords = [x(:); y(:)];

end

function [lmah, rmah] = mean_arc_height(arc, x1, x2, x3, x4, t)

y1 = arc(x1);
y2 = arc(x2);
y3 = arc(x3);
y4 = arc(x4);

% compute angle from slope. fetch boundary cases
% check if left case
if abs(x2 - x1) < eps
    alpha23 = atan((y3-y2)/(x3-x2));
    alpha12 = alpha23;
    alpha34 = atan((y4-y3)/(x4-x3));
% check if right case
elseif abs(x3 - x4) < eps
    alpha12 = atan((y2-y1)/(x2-x1));
    alpha23 = atan((y3-y2)/(x3-x2));
    alpha34 = alpha23;
% else
else
    alpha12 = atan((y2-y1)/(x2-x1));
    alpha23 = atan((y3-y2)/(x3-x2));
    alpha34 = atan((y4-y3)/(x4-x3));
end

% compute arc height
ah12 = t/2 / cos(alpha12);
ah23 = t/2 / cos(alpha23);
ah34 = t/2 / cos(alpha34);

% compute left and right mean arc hight
lmah = (ah12 + ah23)/2;
rmah = (ah23 + ah34)/2;

end

function coords = construct_ldiag(xl, yl, xr, yr, alpha, ts)

% compute angle of diag
beta  = atan((xr-xl)/(yr-yl));
gamma = asin(ts/norm([yl-yr; xl-xr]));
delta = beta - gamma;

% compute x coordinate of upper left point
xu = xr - ts/cos(delta);

% compute dummy value for down right point
phi = delta + alpha;
tmp = ts / cos(phi);
xd  = xl + cos(alpha)*tmp;
yd  = yl + sin(alpha)*tmp;

x = [xl, xu, xr, xd];
y = [yl, yr, yr, yd];

coords = [x(:); y(:)];

end

function coords = construct_rdiag(xr, yr, xl, yl, alpha, ts)

% compute angle of diag
beta  = atan((xl-xr)/(yl-yr));
gamma = asin(ts/norm([yr-yl; xr-xl]));
delta = beta - gamma;

% compute x coordinate of upper left point
xu = xl + ts/cos(delta);

% compute dummy value for down right point
phi = delta + alpha;
tmp = ts / cos(phi);
xd  = xr - cos(alpha)*tmp;
yd  = yr - sin(alpha)*tmp;

x = [xr, xu, xl, xd];
y = [yr, yl, yl, yd];

coords = [x(:); y(:)];

end

function coords = construct_lcol(ldiag, rdiag, ts)

yup = ldiag(6+2);

hr_lcol = yup - (ldiag(2+2)-ldiag(2+1)-ts/2)/ ...
    (ldiag(2+2) - ldiag(2+1)) * (ldiag(6+2) - ldiag(6+1));
if ~isempty(rdiag)
    hl_lcol = yup + (ldiag(2+1)-rdiag(2+2)-ts/2)/ ...
        (rdiag(2+2) - rdiag(2+1)) * (rdiag(6+2) - rdiag(6+1));
else
    hl_lcol = hr_lcol;
end

xl = ldiag(2+1) - ts/2;
xr = ldiag(2+1) + ts/2;

x = [xl;  xl;      xr;      xr];
y = [yup; hl_lcol; hr_lcol; yup];

coords = [x(:); y(:)];
    
end
