function symplectic_experiments( input_params )
%function symplectic_experiments( input_params )
% Script to run the different experiments of the
%   mdpi paper "Symplectic Model Order Reduction with Non-Orthonormal Bases"
%
% The parameter 'input_params' allow selecting specific simulations.
% Default values apply to the experiments used in the paper.
%
% The results are written to several mat files in the folder
% "symplectic_experiments"

if nargin < 1
    input_params = struct();
end

% set default input params
default_input_params = struct();
% model
default_input_params.models          = {
                                        'beam_autonomous' ...
                                        'beam_non_autonomous' ...
                                       };
% select sampling mode
default_input_params.sampling_modes  = {
                                        'generalization'
                                       };
% select integrator
default_input_params.integrators     = {
                                        'implicit_midpoint' ...
                                       };
% select MOR technique
default_input_params.MOR_techniques  = {...
                                        'POD_full_state'... 
                                        'POD_separate_states'... 
                                        'PSD_complex_SVD'...
                                        'PSD_greedy' ...
                                        'PSD_SVD_like_decomp'...
};
% different sizes of the RB to be computed
default_input_params.RB_sizes        = 20:20:300;
% parameter mu to be computed in the reduced simulation
default_input_params.mu_test_repro   = [122e9; 81e9; 7856];
% set time interval for evaluation
default_input_params.tspan_test      = [0,0.072]; % one oscillation of the beam model
% set parameters for generalization case
default_input_params.mu_ranges       = {[35e9,125e9],[35e9,83e9],[7856,7856]};
default_input_params.n_sam_pp        = [3,3,1]; % samples per parameter! (i.e. 9 parameters in total)
% compute samples for generalization case
default_input_params.mu_test_general = compute_mu_test_general(default_input_params);
% other time parameters like dt_test, tspan_train, dt_train, dt_reference
% are set in update_model_specific_input_params

% copy default values to input_params
input_params = set_input_params(default_input_params, input_params);

% these experiments use NONDIMENSIONALISATION!
input_params.use_nondim = true;

% sort RB_sizes
input_params.RB_sizes = sort(input_params.RB_sizes, 'ascend');

% basic file paths 
%   do not change (these paths are also needed in plot script)
base_dir            = 'symplectic_experiments';
experiment_base_str = '%s_%s';
results_file_str    = fullfile(base_dir, '%s_results.mat');
ddate_file_str      = fullfile(base_dir, '%s_detailed_data.mat');
snapshots_file_str  = fullfile(base_dir, '%s_snapshots.mat');
full_svd_file_str   = fullfile(base_dir, '%s_full_svd.mat');
sep_svd_file_str    = fullfile(base_dir, '%s_separate_svd.mat');
svd_like_file_str   = fullfile(base_dir, '%s_svd_like_decomp.mat');
pth_management      = fullfile(base_dir, 'all_experiments.mat');

% generate results dir
if ~(exist(base_dir, 'dir') == 7)
    mkdir(base_dir);
end

% struct to collect all computed experiments in the managment file
experiments = struct('id'           , 0, ...
                     'model'        , 0, ...
                     'sampling_mode', 0);

% cycle all models, sampling modes and integrators
n_models = numel(input_params.models);
i_experiment = 0;
for i_model = 1:n_models
    % set model name
    model_name = input_params.models{i_model};

    % update model specific parameters (msp)
    msp_input_params = update_model_specific_input_params(default_input_params, input_params, model_name);

    % generate specific model
    switch(model_name)
        case 'beam_autonomous'
            model = beam_autonomous_model(msp_input_params);
            param_dep_inital_value = false; % parameter-dependent initial value
        case 'beam_non_autonomous'
            model = beam_non_autonomous_model(msp_input_params);
            param_dep_inital_value = false; % parameter-dependent initial value
        otherwise
            error('Unknown model name %s', model_name);
    end
    
    % set evaluate time
    model = model.set_tspan(model, msp_input_params.tspan_test);
    model = model.set_dt(model, msp_input_params.dt_test);
    
    % generate model data and fetch mass matrix
    model_data = gen_model_data(model);
    M          = model.get_mass_matrix(model, model_data);
    
    n_smodes = numel(msp_input_params.sampling_modes);
    for i_sampling_mode = 1:n_smodes
        % select sampling mode
        sampling_mode = msp_input_params.sampling_modes{i_sampling_mode};
    
        switch(sampling_mode)
            case 'reproduction'
                % change to 'reproduction' case
                model.mu_ranges  = {msp_input_params.mu_test_repro(1)*[1,1],...
                                    msp_input_params.mu_test_repro(2)*[1,1],...
                                    msp_input_params.mu_test_repro(3)*[1,1]};
                model            = model.create_mu_list(model, [1,1,1]);
                mus_test         = msp_input_params.mu_test_repro(:);
            case 'generalization'
                % set RB mu list to 'generalization' case
                model.mu_ranges           = msp_input_params.mu_ranges;
                model.n_samples_per_param = msp_input_params.n_sam_pp;
                model                     = model.create_mu_list(model, model.n_samples_per_param);
                mus_test                  = msp_input_params.mu_test_general;
            otherwise
                error('Unknown sampling mode %s', sampling_mode);
        end
        
        % set snapshots and model name
        experiment = sprintf(experiment_base_str, model_name, sampling_mode);
        model.pth_snapshots = sprintf(snapshots_file_str, experiment);
        model.pth_full_svd  = sprintf(full_svd_file_str, experiment);
        model.pth_sep_svd   = sprintf(sep_svd_file_str, experiment);
        model.pth_svd_like  = sprintf(svd_like_file_str, experiment);
        
        % number of different variables
        n_techs       = numel(msp_input_params.MOR_techniques);
        n_RB_sizes    = numel(msp_input_params.RB_sizes);
        n_ints        = numel(msp_input_params.integrators);
        n_mus_test    = size(mus_test, 2);
        % compute number of time steps
        n_ts          = 1+round(diff(msp_input_params.tspan_test)/msp_input_params.dt_test);
        % estimate number of entries in table
        n_entries     = n_techs * n_RB_sizes * n_ints * n_mus_test;

        % save input parameters in results
        results              = struct();
        results.input_params = msp_input_params;
        results.mus_test     = mus_test;

        % save model parameters in results
        results.model = model;

        % allocate arrays for timetables for time-dependent results
        err_L2_id    = zeros(n_ts, n_entries);
        err_Linf_id  = zeros(n_ts, n_entries);
        Ham_id       = zeros(n_ts, n_entries);
        HamKin_id    = zeros(n_ts, n_entries);
        HamPot_id    = zeros(n_ts, n_entries);
        % allocate struct for entries in the table
        table_data   = struct('id'               , 0,  ...
                              'MOR_technique'    , cell(n_entries,1), ...
                              'RB_size'          , 0, ...
                              'actual_RB_size'   , 0, ...
                              'integrator'       , cell(n_entries,1), ...
                              'lame_lambda'      , 0, ...
                              'lame_mu'          , 0, ...
                              'mean_rel_err_Linf', 0, ...
                              'max_rel_err_Linf' , 0, ...
                              'max_rel_diff_Ham' , 0, ...
                              'mean_rel_diff_Ham', 0);
        i_entry    = 1;

        % table for RB data
        n_RB_entries  = n_techs*n_RB_sizes*n_ints;
        RB_table_data = struct('MOR_technique'     , cell(n_RB_entries,1), ...
                               'RB_size'           , 0, ...
                               'actual_RB_size'    , 0, ...
                               'proj_err'          , 0, ...
                               'orthogonality'     , 0, ...
                               'J_orthogonality'   , 0);
        i_RB_entry = 1;

        % allocate path to save results mat
        pth_results_mat = sprintf(results_file_str, experiment);
        if (exist(pth_results_mat,'file') == 2)
            msg = sprintf('The experiment for model: ''%s'', samling mode: ''%s'' does already exist! Do you want to overwrite? Y/N [N]: ', model_name, sampling_mode);
            ch  = input(msg, 's');
            if ~strcmpi(ch, 'y') && ~strcmpi(ch, 'yes')
                fprintf('You chose to not overwrite the experiment for model: ''%s'', samling mode: ''%s''. Execution is aborted!\n', model_name, sampling_mode);
                continue;
            end
        end
        
        % set evaluate time
        model = model.set_tspan(model, msp_input_params.tspan_test);
        model = model.set_dt(model, msp_input_params.dt_test);
        
        % load detailed data if available
        pth_ddata_mat = sprintf(ddate_file_str, experiment);
        if (exist(pth_ddata_mat,'file') == 2)
            b_compute_ddata = false(n_mus_test, 1);
            % load detailed data
            load(pth_ddata_mat);
        else
            b_compute_ddata = true(n_mus_test, 1);
            n = size(M, 1);
            ddata = struct('mu'       , zeros(3, n_mus_test), ...
                           'dofs'     , zeros(2*n, n_ts, n_mus_test), ...
                           'Ham'      , zeros(n_ts, n_mus_test), ...
                           'REL_diff_Ham'   , zeros(n_mus_test,1), ...
                           'REL_Err'  , zeros(n_mus_test,1));
        end

        % loop all techniques with different RB_sizes
        for i_tech = 1:n_techs
            MOR_technique = msp_input_params.MOR_techniques{i_tech};

            % choose MOR technique
            model.MOR_technique = MOR_technique;

            for i_RB_sizes = 1:n_RB_sizes
                RB_size = msp_input_params.RB_sizes(i_RB_sizes);

                % check if RB size is valid
                %   care! POD only takes half the sizes of PSD
                MOR_principle = strsplit(MOR_technique, '_');
                MOR_principle = MOR_principle{1};
                if any(strcmp(MOR_principle, {'POD', 'PH'}))
                    max_RB_size = n_ts * n_mus_test;
                elseif any(strcmp(MOR_principle, {'PSD', 'Jorthonormalized'}))
                    max_RB_size = 2*n_ts * n_mus_test;
                else
                    error('Unknown MOR principle: %s', MOR_principle);
                end

                if RB_size > max_RB_size
                    continue;
                end

                % set training time
                model = model.set_tspan(model, msp_input_params.tspan_train);
                model = model.set_dt(model, msp_input_params.dt_train);
                % set integrator
                model.integrator = 'rescaled_implicit_midpoint';

                % Choose RB size
                model.RB_size = RB_size;
                % Generate RB
                try
                    detailed_data = gen_detailed_data(model,model_data);
                    
                % Catch error if basis size does not match
                %   this might occur in the greedy, if RB_size > rank(snapshot_matrix)
                catch ME
                    if (strcmp(ME.identifier, 'RBmatlab:DynLinElast:GenDetailedData:RBSizeMissmatch'))
                        % break the loop (RB_sizes are supposed to be sorted)
                        break;
                    else
                        rethrow(ME)
                    end
                end

                for i_mu_test = 1:n_mus_test
                    mu_test = mus_test(:,i_mu_test);

                    % set mu for detailed reference solution
                    model = set_mu(model, mu_test);

                    % compute weighting matrix
                    K = model_data.operators(model, model_data);
                    H = @(x) [K*x(1:end/2,:); M\x(end/2+1:end,:)];

                    % compute or load reference solution
                    if b_compute_ddata(i_mu_test)
                        % set evaluate time
                        model = model.set_tspan(model, msp_input_params.tspan_test);
                        model = model.set_dt(model, msp_input_params.dt_reference);
                        % set integrator
                        model.integrator = 'rescaled_implicit_midpoint';

                        % choose special time steps
                        idx_t = (0:n_ts-1)*msp_input_params.dt_ref_multiple+1;
                        % do detailed simulation with same mu as reduced
                        sim_data = detailed_simulation(model, model_data);
                        % sanity check for time step selection
                        if sim_data.t(idx_t) - (0:n_ts-1)*msp_input_params.dt_test > eps
                            error('time step selection for reference solution is wrong!');
                        end
                        % save 
                        dofs                      = sim_data.uh.dofs;
                        Ham                       = sim_data.Ham;
                        ddata.mu(:,i_mu_test)     = mu_test;
                        ddata.dofs(:,:,i_mu_test) = dofs(:,idx_t);
                        ddata.Ham(:,i_mu_test)    = Ham(idx_t);
                        ddata.REL_diff_Ham(i_mu_test)   = ...
                            max( max(abs(sim_data.HamKin(idx_t) - sim_data.HamKin(idx_t(1)))), ...
                                 max(abs(sim_data.HamPot(idx_t) - sim_data.HamPot(idx_t(1)))) );
                        ddata.REL_Err(i_mu_test)   = ...
                            max(max(abs(dofs), [], 1));

                        REL_diff_Ham = ddata.REL_diff_Ham(i_mu_test);
                        REL_Err = ddata.REL_Err(i_mu_test);

                        % select time steps for reference solution
                        sim_data.uh.dofs = dofs(:,idx_t);
                        sim_data.Ham     = ddata.Ham(:,i_mu_test);

                        b_compute_ddata(i_mu_test) = false;
                    else
                        if norm(ddata.mu(:,i_mu_test) - mu_test) > 0
                            error('Detailed data does not match the desired format!');
                        end

                        uh           = Fem.DiscFunc(ddata.dofs(:,:,i_mu_test), model_data.truncated_df_info);
                        sim_data     = struct();
                        sim_data.uh  = uh;
                        sim_data.Ham = ddata.Ham(:,i_mu_test);

                        REL_diff_Ham = ddata.REL_diff_Ham(i_mu_test);
                        REL_Err      = ddata.REL_Err(i_mu_test);
                    end

                    % set evaluate time
                    model = model.set_tspan(model, msp_input_params.tspan_test);
                    model = model.set_dt(model, msp_input_params.dt_test);

                    % create reduced basis in reduced data
                    reduced_data = gen_reduced_data(model,detailed_data);

                    for i_int=1:n_ints
                        integrator = msp_input_params.integrators{i_int};

                        switch(integrator)
                            case 'implicit_midpoint'
                                model.integrator = 'rescaled_implicit_midpoint';
                            case 'backward_Euler'
                                model.integrator = 'rescaled_implicit_euler';
                            otherwise
                                error('Unknown integrator: %s', integrator);
                        end

                        % add entry to table data
                        table_data(i_entry).id             = i_entry;
                        table_data(i_entry).MOR_technique  = MOR_technique;
                        table_data(i_entry).integrator     = integrator;
                        table_data(i_entry).RB_size        = RB_size;
                        table_data(i_entry).actual_RB_size = detailed_data.actual_RB_size;
                        table_data(i_entry).lame_lambda    = mu_test(1);
                        table_data(i_entry).lame_mu        = mu_test(2);

                        % run reduced simulation
                        rb_sim_data = rb_simulation(model,reduced_data);
                        rb_sim_data = rb_reconstruction(model,detailed_data,rb_sim_data);

                        % evaluate errors
                        err_L2_id(:, i_entry)   = eval_L2_err(rb_sim_data, sim_data);
                        err_Linf                = eval_Linf_err(rb_sim_data, sim_data);
                        err_Linf_id(:, i_entry) = err_Linf;
                        
                        % evalute difference in the Hamiltonian
                        diff_Ham = eval_diff_Ham(rb_sim_data, sim_data);

                        % save Hamiltonian
                        Ham                   = rb_sim_data.Ham;
                        Ham_id(:, i_entry)    = Ham;
                        HamKin_id(:, i_entry) = rb_sim_data.HamKin;
                        HamPot_id(:, i_entry) = rb_sim_data.HamPot;

                        % add computable data to table
                        table_data(i_entry).rel_err_Linf  = max(err_Linf) / REL_Err;
                        table_data(i_entry).mean_rel_diff_Ham = mean(diff_Ham) / REL_diff_Ham;
                        table_data(i_entry).max_rel_diff_Ham  = max(diff_Ham) / REL_diff_Ham;
                        
                        % increment entry index
                        i_entry = i_entry + 1;
                    end

                end % end mu_test

                if ~param_dep_inital_value
                    if i_RB_sizes==1
                        % load snapshots for projection error
                        load(model.pth_snapshots, 'all_dofs');
                    end

                    RB_table_data(i_RB_entry).MOR_technique   = MOR_technique;
                    RB_table_data(i_RB_entry).RB_size         = RB_size;
                    RB_table_data(i_RB_entry).actual_RB_size  = detailed_data.actual_RB_size;
                    RB_table_data(i_RB_entry).proj_err        = eval_proj_err(all_dofs, detailed_data.RB, detailed_data.W);
                    RB_table_data(i_RB_entry).orthogonality   = eval_orthogonality(detailed_data.RB);
                    RB_table_data(i_RB_entry).J_orthogonality = eval_J_orthogonality(detailed_data.RB);
                    i_RB_entry = i_RB_entry + 1;
                end

            end % end RB sizes
        end
    
        % save detailed data as mat file
        if ~(exist(pth_ddata_mat,'file') == 2)
            save(pth_ddata_mat, 'ddata');
        end
        
        % assemble results table
        res_table = struct2table(table_data);
        % delete empty rows
        if n_entries >= i_entry
            fprintf('res table conversion: deleting %d empty rows\n', n_entries-i_entry+1);
            res_table(i_entry:end,:)   = [];
            % truncate timetable data
            err_L2_id(:,i_entry:end)   = [];
            err_Linf_id(:,i_entry:end) = [];
            Ham_id(:,i_entry:end)      = [];
            HamKin_id(:,i_entry:end)   = [];
            HamPot_id(:,i_entry:end)   = [];
        else
            fprintf('res table conversion: added %d rows in total\n', i_entry-n_entries-1);
        end
        % save in results
        results.res_table = res_table;

        % assemble RB table
        RB_table = struct2table(RB_table_data);
        % delete empty rows
        if n_RB_entries >= i_RB_entry
            fprintf('RB table conversion: deleting %d empty rows\n', n_RB_entries-i_RB_entry+1);
            RB_table(i_RB_entry:end,:)   = [];
        else
            fprintf('RB table conversion: added %d rows in total\n', i_RB_entry-n_RB_entries-1);
        end
        % save in results
        results.RB_table = RB_table;

        % assemble timetables
        rowTimes         = seconds(model.tspan(1) + (0:n_ts-1) * model.dt);
        id               = err_L2_id;
        results.err_L2   = array2timetable(id,'RowTimes',rowTimes);
        id               = err_Linf_id;
        results.err_Linf = array2timetable(id,'RowTimes',rowTimes);
        id               = Ham_id;
        results.Ham      = array2timetable(id,'RowTimes',rowTimes);
        id               = HamKin_id;
        results.HamKin   = array2timetable(id,'RowTimes',rowTimes);
        id               = HamPot_id;
        results.HamPot   = array2timetable(id,'RowTimes',rowTimes);

        % save results as mat file
        save(pth_results_mat, 'results');

        % update mangement file
        i_experiment = i_experiment + 1;
        experiments(i_experiment).id = i_experiment;
        experiments(i_experiment).model_name = model_name;
        experiments(i_experiment).sampling_mode = sampling_mode;
        save(pth_management, 'experiments');
    end
end
    
end

function mu_test = compute_mu_test_general(default_input_params)

n_s      = max(default_input_params.n_sam_pp - 1,1); % number of sections 
n_sam_ps = 4; % samples per section
mu_test  = zeros(3, prod(n_s) * n_sam_ps);
% seeds the random number generator using the nonnegative
% integer seed so that rand, randi, and randn produce a
% predictable sequence of number
rng(42);

offset1 = default_input_params.mu_ranges{1}(1);
offset2 = default_input_params.mu_ranges{2}(1);
diff1   = diff(default_input_params.mu_ranges{1}) / n_s(1);
diff2   = diff(default_input_params.mu_ranges{2}) / n_s(2);
for ii=1:n_s(1)
    c_offset1 = offset1 + (ii-1) * diff1;
    for jj=1:n_s(2)
        c_offset2 = offset2 + (jj-1) * diff2;

        i_test = (ii-1)*n_s(1)*n_sam_ps + ...
            (jj-1)*n_sam_ps + (1:n_sam_ps);
        mu_test(1,i_test) = c_offset1 + rand(1,n_sam_ps) * diff1;
        mu_test(2,i_test) = c_offset2 + rand(1,n_sam_ps) * diff2;
        mu_test(3,i_test) = 7856;
    end
end
    
end

% set input parameters
function input_params = set_input_params(default_input_params, input_params, names)

% apply all default options if not specified otherwise
if nargin < 3
    names = fieldnames(default_input_params);
end

if ~iscell(names)
    names = {names};
end

for i_n = 1:numel(names)
    name = names{i_n};

    if isfield(default_input_params, name)
        options = default_input_params.(name);

        if ~isfield(input_params, name)
            input_params.(name) = options;
        else
            % check if current option is valid option
            %   needs only to be checked for char options
            if iscell(options) && ischar(options{1})
                option = input_params.(name);
                if ischar(option)
                    option = {option};
                    input_params.(name) = option;
                end
                for i_option=1:numel(option)
                    if ~ismember(option{i_option}, options)
                        error('Unknown value for %s: %s', name, option{i_option});
                    end
                end
            end
        end
    end
end

end

% update model specific input parameters
function msp_input_params = update_model_specific_input_params(default_input_params, input_params, model)

% copy input parameters
msp_input_params = input_params;

% sanity check
if ~ismember(model, msp_input_params.models)
    error(['Cannot update model parameters for the model (%s) since',...
        ' it is not contained in the models in the input parameters!']);
end

if isfield(msp_input_params, model)
    model_spec_params = msp_input_params.(model);
    msp_names = fieldnames(model_spec_params);
    
    for i_msp = 1:numel(msp_names)
        key = msp_names{i_msp};
        model_spec_params = set_input_params(default_input_params, model_spec_params, key);
        msp_input_params.(key) = model_spec_params.(key);
    end
end

% finally, update paramters for time
% set time data for test
if isfield(msp_input_params, 'tspan_test') && ~isfield(msp_input_params, 'dt_test')
    msp_input_params.dt_test = diff(msp_input_params.tspan_test)/150;
end
% set time interval for training
if isfield(msp_input_params, 'tspan_test') && ~isfield(msp_input_params, 'tspan_train')
    msp_input_params.tspan_train = msp_input_params.tspan_test;
end
if isfield(msp_input_params, 'dt_test') && ~isfield(msp_input_params, 'dt_train')
    msp_input_params.dt_train = msp_input_params.dt_test;
end

% set the time step for the computation of the refrence solution
if isfield(msp_input_params, 'dt_test') && ~isfield(msp_input_params, 'dt_reference')
    msp_input_params.dt_reference = msp_input_params.dt_test;
end
% sanity check for time step for reference solution
dt_ref_multiple = msp_input_params.dt_test/msp_input_params.dt_reference;
if dt_ref_multiple - round(dt_ref_multiple) > eps
    error(['the time step for the computation of the reference solution',...
        ' (dt_reference) needs to be an integral multiple of dt_test']);
end
msp_input_params.dt_ref_multiple = round(dt_ref_multiple);

end

% evaluate orthogonailty of the reduced basis RB
function orth = eval_orthogonality(RB)

k   = size(RB,2) / 2;
I2k = speye(2*k, 2*k); 

orth = norm(RB'*RB - I2k, 'fro');

end

% evaluate J-orthogonailty of the reduced basis RB
function Jorth = eval_J_orthogonality(RB)

k   = size(RB,2) / 2;
n   = size(RB,1) / 2;
I2k = speye(2*k, 2*k);
J2n = [sparse(n, n), speye(n, n); ...
       -speye(n, n), sparse(n, n)];
J2k = [sparse(k, k), speye(k, k); ...
       -speye(k, k), sparse(k, k)];

Jorth = norm(J2k'*RB'*J2n*RB - I2k, 'fro');

end

% compute projection error of RB V with projection matrix W wrt to the
% snapshot matrix Xs
function proj_err = eval_proj_err(Xs, V, W)

s  = size(V,1);
Is = speye(s,s);

proj_err = sum(sum(((Is - V*W')*Xs).^2,1));

end

% evaluate L_inf error
function err = eval_L2_err(rb_sim_data, sim_data)

x     = sim_data.uh.dofs;
Vxhat = rb_sim_data.uh.dofs;

err = x - Vxhat;
err = sqrt(sum(err.^2, 1));

end

% evaluate L_inf error
function err = eval_Linf_err(rb_sim_data, sim_data)

x     = sim_data.uh.dofs;
Vxhat = rb_sim_data.uh.dofs;

err = x - Vxhat;
err = max(abs(err), [], 1);

end

% evaluate difference in the Hamiltonian
function diff_Ham = eval_diff_Ham(rb_sim_data, sim_data)

Ham = sim_data.Ham;
redHam = rb_sim_data.Ham;

diff_Ham = abs((Ham(:)-Ham(1)) - (redHam(:)-redHam(1)));

end

