function symplectic_plot_results( plot_params )
%function symplectic_plot_results( plot_params )
% This script plots the results of the script
%   symplectic_experiments(input_params)
% The plots are only available if the results are available as .mat file on
%   the Matlab search path. It is not checked if the .mat files are available
%   so this function might crash due to the lack of the specific .mat file.
% Running symplectic_experiments() without input parameters
%   in advance of executing this script creates all .mat files required.

if nargin < 1
    plot_params = struct();
end

if ~isfield(plot_params, 'tol_plot')
    plot_params.tol_plot = 50;
end

if ~isfield(plot_params, 'plot_error_logarithmic')
    plot_params.plot_error_logarithmic = true;
end

% tolerance for Hamiltonian
%   to determine if Hamiltonian is preserves
if ~isfield(plot_params, 'tolHam')
    plot_params.tolHam = 1e-10;
end

% find management mat
base_dir            = 'symplectic_experiments';
experiment_base_str = '%s_%s';
results_file_str    = fullfile(base_dir, '%s_results.mat');
ddate_file_str      = fullfile(base_dir, '%s_detailed_data.mat');
snapshots_file_str  = fullfile(base_dir, '%s_snapshots.mat');
full_svd_file_str   = fullfile(base_dir, '%s_full_svd.mat');
sep_svd_file_str    = fullfile(base_dir, '%s_separate_svd.mat');
svd_like_file_str   = fullfile(base_dir, '%s_svd_like_decomp.mat');
pth_management      = fullfile(base_dir, 'all_experiments.mat');
% load mangement file
load(pth_management, 'experiments');

n_experiments = numel(experiments);
for i_experiment=1:n_experiments
    model_name = experiments(i_experiment).model_name;
    sampling_mode = experiments(i_experiment).sampling_mode;
    
    % Load results
    experiment = sprintf(experiment_base_str, model_name, sampling_mode);
    load(sprintf(results_file_str, experiment), 'results');
    exp_title = [strip_underscore(model_name), ', ', strip_underscore(sampling_mode)];

    % Copy data from results
    if ~isfield(plot_params, 'MOR_techniques')
        plot_params.MOR_techniques = results.input_params.MOR_techniques;
    end

    if ~isfield(plot_params, 'integrators')
        plot_params.integrators = results.input_params.integrators;
    end

    if ~isfield(plot_params, 'RB_sizes')
        plot_params.RB_sizes = results.input_params.RB_sizes;
    end

    if ~isfield(plot_params, 'mu_test_general')
        plot_params.mu_test_general = results.input_params.mu_test_general;
    end

    if ~isfield(plot_params, 'mu_test_repro')
        plot_params.mu_test_repro = results.input_params.mu_test_repro;
    end

    %% Detailed over time plots
    plot_over_time(plot_params, results, {'err_Linf'}, 'YScale', 'log');
    title(exp_title);
    xlabel('number i of the time step t_i');
    ylabel('Error in semilogy');
    plot_over_time(plot_params, results, {'Ham', 'HamKin', 'HamPot'});
    title(exp_title);
    xlabel('number i of the time step t_i');
    ylabel('Hamiltonian and its parts');

    keyboard();
    close all;

    %% plot L2 projection error
    figure();
    plot_params.sampling_mode = sampling_mode;
    multiple_plot_over_RB_size(plot_params, results.RB_table, 'proj_err');
    title(exp_title);
    xlabel('size of the ROB');
    ylabel('L2 projection error');

    keyboard();
    close all;

    %% plot J orthognoality
    figure();
    plot_params.sampling_mode = sampling_mode;
    multiple_plot_over_RB_size(plot_params, results.RB_table, 'J_orthogonality');
    title(exp_title);
    xlabel('size of the ROB');
    ylabel('J orthogonality');
    
    keyboard();
    close all;

    %% plot orthogonality
    figure();
    plot_params.sampling_mode = sampling_mode;
    multiple_plot_over_RB_size(plot_params, results.RB_table, 'orthogonality');
    title(exp_title);
    xlabel('size of the ROB');
    ylabel('orthogonality');

    keyboard();
    close all;
    
    %% heatmap plots
    res_tab = results.res_table;
    res_tab.preserves_Hamiltonian = res_tab.max_rel_diff_Ham < plot_params.tolHam;
    % plot heatmap for repro
    figure;
    heatmap(res_tab, 'MOR_technique', 'integrator', ...
        'ColorVariable', 'preserves_Hamiltonian', 'ColorMethod', 'mean', ...
        'Title', [exp_title, ': mean of preserves_Hamiltonian']);
    
    keyboard();
    close all;
    
end

end

% Compressed plots (no time axis)
function multiple_plot_over_RB_size(params, table, field)

MOR_techs = params.MOR_techniques;

n_techs = numel(MOR_techs);

%colors = get(groot,'defaultAxesColorOrder');
colors = [
         0    0.4470    0.7410
    0.8500    0.3250    0.0980
    0.9290    0.6940    0.1250
    0.4940    0.1840    0.5560
    0.4660    0.6740    0.1880
    0.3010    0.7450    0.9330
    0.6350    0.0780    0.1840
    0.0000    0.4470    0.74100
    ];
i_colors = 1;
n_colors = size(colors, 1);

h_plot = zeros(n_techs,1);

leg_entries = cell(1,n_techs);

hold on;
for i_techs = 1:n_techs
    tech = MOR_techs{i_techs};
    params.MOR_technique = tech;

    % Plot mean values
    h_plot(i_techs) = plot_over_RB_size(params, table, field, ...
        'Marker', 'x', 'LineWidth', 2, 'Color', colors(mod(i_colors-1, n_colors)+1,:), 'LineStyle', 'none');

    leg_entries{i_techs} = strip_underscore(tech);
    i_colors = i_colors + 1;
end
hold off;

% set logartihmic y scale if desired
if params.plot_error_logarithmic
    set(gca, 'YScale', 'log');
end

if numel(h_plot) > 1
    legend(h_plot, leg_entries);
end

end

function h = plot_over_RB_size(params, table, field, varargin)

MOR_tech = params.MOR_technique;

% mask data from table
mask = strcmp(table.MOR_technique, MOR_tech);

if ismember('integrator', table.Properties.VariableNames)
    int  = params.integrator;
    mask = mask & strcmp(table.integrator, int);
end

if isfield(params, 'boxplot') && params.boxplot
    n_mu = size(params.mu_test_general, 2);
    x = unique(table(mask, :).RB_size);
    bp_data = zeros(n_mu, numel(x));
    % plot single boxplots (analogously to PGFplots)
    for i_x = 1:numel(x)
        maskx          = mask & (table.RB_size == x(i_x));
        bp_data(:,i_x) = table(maskx, :).(field);
    end
    boxplot(bp_data, 'Labels', num2str(x));
    h = 0;
else
    x = table(mask, :).RB_size;
    y = table(mask, :).(field);
    h = plot(x,y,varargin{:});
end
     
end

% Over time plots
function plot_over_time(params, results, fields, varargin)
% see https://de.mathworks.com/help/control/ug/build-app-with-interactive-plot-updates.html

MOR_techs  = params.MOR_techniques;
ints       = params.integrators;
RB_sizes   = params.RB_sizes;
mu_test    = results.mus_test;

model = results.model;

nt = round(diff(model.tspan)/model.dt);
t  = model.tspan(1) + (0:nt)*model.dt;

f = figure();
fpos = get(f, 'Position');
set(f, 'Units', 'pixels', 'Position', [fpos(1:2), 560, 500]);
set(f, 'Units', 'normalized');
ax = axes('Parent',f,'Position',[0.15 0.48  0.8 0.45]);
if iscell(fields)
    y = zeros(numel(t), numel(fields));
    id = get_id_from_table(results.res_table, MOR_techs{1}, ints{1}, RB_sizes(1), mu_test(:,1));
    for i_fields = 1:numel(fields)
        y(:, i_fields) = results.(fields{i_fields}).(['id', num2str(id)]);
    end
    plot(ax, t, y);
    legend(strip_underscore(fields));
else
    error('Not implemented');
end
set(ax,  'Tag', 'b_ax', 'NextPlot', 'replacechildren', varargin{:});

up = @(b, bl) update_plot_over_time(f, t, results, fields, b, bl);
my_slider(f, 'b_tech' , [0.1 0.28  0.8 0.04], MOR_techs, 'MOR technique: ', up);
my_slider(f, 'b_int'  , [0.1 0.20  0.8 0.04], ints, 'integrator: ', up);
my_slider(f, 'b_RBs'  , [0.1 0.12  0.8 0.04], RB_sizes, 'RB size: 2k=', up);
my_slider(f, 'b_mu'  ,  [0.1 0.04  0.8 0.04], make_paramTest_string(mu_test), '', up);
set(f, 'UserData', mu_test);
end

function update_plot_over_time(f, t, results, fields, b, bl)

% check if current value is 
v = round(get(b,'Value'));
if v < get(b, 'Min') || v > get(b, 'Max')
    set(b,'Value', get(b, 'Min'))
    return;
end
set(b,'Value',v);

% update current label
labels = get(b, 'UserData');
if iscell(labels)
    label = strip_underscore(labels{v});
elseif isvector(labels)
    label = num2str(labels(v));
else
    error('Variable labels is of unknown type.');
end
base_str = get(bl, 'UserData');
set(bl, 'String', [base_str, label]);

% find current options
b_tech = findobj(f, 'Tag', 'b_tech');
i_tech = get(b_tech, 'Value');
tech   = get(b_tech, 'UserData');
tech   = tech{i_tech};

b_int = findobj(f, 'Tag', 'b_int');
i_int = get(b_int, 'Value');
int   = get(b_int, 'UserData');
int   = int{i_int};

b_RBs = findobj(f, 'Tag', 'b_RBs');
i_RBs = get(b_RBs, 'Value');
RBs   = get(b_RBs, 'UserData');
RBs   = RBs(i_RBs);


mu_slider = 'b_mu';
b_mu      = findobj(f, 'Tag', mu_slider);
i_mu = get(b_mu, 'Value');
mu   = get(f, 'UserData');
mu   = mu(:,i_mu);

% plot current result for current options
ax = findobj(f, 'Tag', 'b_ax');
if iscell(fields)
    y = zeros(numel(t), numel(fields));
    id = get_id_from_table(results.res_table, tech, int, RBs, mu);
    if ~isempty(id)
        for i_fields = 1:numel(fields)
            y(:, i_fields) = results.(fields{i_fields}).(['id', num2str(id)]);
        end
        plot(ax, t, y);
        legend(fields{:});
    else
        cla(ax);
    end
else
    error('Not implemented');
end

end

function b = my_slider(f, silder_name, pos, labels, base_str, up)

n_steps = numel(labels);
h = pos(4);
bgcolor = f.Color;

% First label
v = 1;
if iscell(labels)
    label = strip_underscore(labels{v});
elseif isvector(labels)
    label = num2str(labels(v));
else
    error('Variable labels is of unknown type.');
end

% create slider and label
b  = uicontrol('Parent',f,'Style','slider','Units', 'normalized','Position',pos,...
               'min',1,'max',n_steps,'SliderStep',[min(1/(n_steps-1),1),1],...
               'Value', v, 'UserData', labels, 'Tag', silder_name);
bl = uicontrol('Parent',f,'Style','text','Units', 'normalized','Position',pos-[0,h,0,0],...
                'String',[base_str,label],'BackgroundColor',bgcolor, 'Tag', ['bl', silder_name(2:end)], ...
                'UserData', base_str);
set(b, 'Callback', @(es,ed) up(b, bl) );

end

function str_paramTest = make_paramTest_string(paramTest)

n_paramTest   = size(paramTest,2);
str_paramTest = cell(n_paramTest, 1);

for i_paramTest = 1:n_paramTest
    str_paramTest{i_paramTest} = sprintf('lambda=%4.2e, mu=%4.2e, rho=%4.2e', ...
        paramTest(:,i_paramTest));
    
end
end

function id = get_id_from_table(...
    table, MOR_technique, integrator, RB_size, mu)

mask = strcmp(table.MOR_technique, MOR_technique) & ...
    strcmp(table.integrator, integrator) & ...
    table.RB_size == RB_size & ...
    table.lame_lambda == mu(1) & ...
    table.lame_mu == mu(2);

masked_table = table(mask, :);
id           = masked_table.id;

end

function strings = strip_underscore(strings)
strings = strrep(strings, '_', ' ');
end
